from tqdm import tqdm
import numpy as np
import scipy.sparse as sp
from scipy.spatial.transform import Rotation
from osqp import OSQP
import math
import warnings

from .base import FrameField

from ...mesh.datatypes import PolyLine,SurfaceMesh,VolumeMesh
from ...mesh.mesh_data import RawMeshData
from ...mesh.mesh_attributes import Attribute, ArrayAttribute

from ... import attributes

from ... import geometry as geom
from ... import utils
from ...geometry import Vec, SphericalHarmonics
from ...geometry.rotations import match_rotation
from ...geometry import transform

from ...operators import laplacian_tetrahedra
from ...procedural import axis_aligned_cube


class FrameField3DCells(FrameField):

    class FileParsingError(Exception):
        def __init__(self, line, message):
            message = f"Reading error at line {line}: {message}"
            super().__init__(message)

    def __init__(self, 
        supporting_mesh : VolumeMesh, 
        verbose=True,
        **kwargs):
        """
        Parameters:
            supporting_mesh (VolumeMesh): the input mesh
            verbose (bool, optional): verbose mode. Defaults to True.
        """
        super().__init__("cells", "FrameField3D", verbose=verbose)
        self.mesh : VolumeMesh = supporting_mesh
        self.mesh.enable_boundary_connectivity()
        self.n_smooth = kwargs.get("n_smooth", 3)
        self.smooth_attach_weight = kwargs.get("smooth_attach_weight", None)
        self._boundary_mesh : SurfaceMesh = None

        self._fnormals : ArrayAttribute = None
        self._cell_on_bnd : Attribute = None

        self.frames : list = None # raw representation of frames as scipy.Rotation objects
        self.var = np.zeros(9*len(self.mesh.cells))

        self._singul_vertices : Attribute = None
        self._singul_edges : Attribute = None

    def read_from_file(self, file_path :str):
        """Reads the values of the frames from a .frame file

        file is supposed to have the following syntax:

            FRAME
            number_of_frames
            a1x a1y a1z b1x b1y b1z c1x c1y c1z
            a2x a2y a2z b2x b2y b2z c2x c2y c2z
            ...
            anx any anz bnx bny bnz cnx cny cnz
            END

        Parameters:
            file_path (str): path to the file

        Raises:
            FrameField3DCells.FileParsingError: if the file is not of the correct format
        """
        if self.initialized:
            warnings.warn("reading frame field from file has erased current frames.")
            self.var = np.zeros(9*len(self.mesh.cells))
        self.frames = []
        with open(file_path, "r") as f:
            data = [l.strip() for l in f.readlines()]
        
        if data[0] != "FRAME":
            # Line 1 should only contain the word "FRAME"
            raise FrameField3DCells.FileParsingError(1,"Invalid frame field file.")
        
        try:
            # Line 2 should only contain the total number of frames
            n_frames = int(data[1])
        except Exception:
            raise FrameField3DCells.FileParsingError(2, f"Invalid number of frames '{data[1]}'")
        
        if n_frames != len(self.mesh.cells):
            raise FrameField3DCells.FileParsingError(2, f"Read number of frames ({n_frames}) does not match number of cells in the mesh ({len(self.mesh.cells)})")
        
        for line in range(n_frames):
            try:
                mat = np.array([float(x) for x in data[line+2].split()]).reshape((3,3)).T
                frame = Rotation.from_matrix(mat)
                self.frames.append(frame)
                self.var[9*line:9*(line+1)] = SphericalHarmonics.from_frame(frame)
            except Exception:
                raise FrameField3DCells.FileParsingError(line+3, f"Invalid frame {data[line+2]}")
        self.initialized = True

    def initialize(self):
        self.frames = [Rotation.identity() for _ in self.mesh.id_cells]
        
        self.log(" | Compute boundary manifold")
        self._boundary_mesh = self.mesh.boundary_connectivity.mesh
        self._fnormals = attributes.face_normals(self._boundary_mesh)
        self._cell_on_bnd = attributes.cell_faces_on_boundary(self.mesh)

        self.log(" | Compute spherical harmonics coordinates on the boundary")
        face_bases_sh = Attribute(float, 9)
        for iF,F in enumerate(self._boundary_mesh.faces):
            NF = self._fnormals[iF]
            axis = geom.cross(Vec(0.,0.,1.), NF)
            if abs(NF.z)<0.99:
                axis = Vec.normalized(axis) * math.atan2(axis.norm(), NF.z)
                face_bases_sh[iF] = SphericalHarmonics.from_vec3(axis)
            else:
                face_bases_sh[iF] = Vec(0., 0., 0., 0., 1., 0., 0., 0., 0.)
        for iFb in self._boundary_mesh.id_faces:
            iF = self.mesh.boundary_connectivity.b2m_face[iFb]
            iC = self.mesh.connectivity.face_to_cells(iF)[0]
            f,a = SphericalHarmonics.project_to_frame(face_bases_sh[iFb])
            self.var[9*iC:9*(iC+1)] += a
            self.frames[iC] *= f
        
        # normalization
        for iC in self._cell_on_bnd:
            nrm = geom.norm(self.var[9*iC:9*(iC+1)])
            if nrm > 1e-8 : self.var[9*iC:9*(iC+1)] /= nrm
        self.initialized = True

    def compute_frame_from_sh(self):
        iterator = self.mesh.id_cells
        if self.verbose: iterator = tqdm(iterator, desc="sh_to_frames")
        for iC in iterator:
            shv = Vec(self.var[9*iC:9*(iC+1)])
            f,a = SphericalHarmonics.project_to_frame(shv)
            self.var[9*iC:9*(iC+1)] = a
            self.frames[iC] = f

    def compute_sh_from_frames(self):
        for i, frame in enumerate(self.frames):
            self.var[9*i:9*(i+1)] = SphericalHarmonics.from_frame(frame)

    def normalize(self):
        """Calls projection for every vertex and also updates representations variables. 
        The projection guarantees that boundary frames stay aligned with their normal.
        """

        for iC in self.mesh.id_cells:
            nrm = geom.norm(self.var[9*iC:9*(iC+1)])
            if nrm>1e-8: 
                self.var[9*iC:9*(iC+1)] /= nrm

        for iC in tqdm(self.mesh.id_cells,desc="normalize1"):
            if self._cell_on_bnd[iC]==0:
                frame, a = SphericalHarmonics.project_to_frame(self.var[9*iC:9*(iC+1)], stop_threshold=1e-3)
                self.var[9*iC:9*(iC+1)] = a
                self.frames[iC] = frame

        for iFb in tqdm(self._boundary_mesh.id_faces, desc="normalize2"):
            iF = self.mesh.boundary_connectivity.b2m_face[iFb]
            iC = self.mesh.connectivity.face_to_cells(iF)[0]
            # if self.cell_on_bnd[iC]!=1 : continue
            nrml = self._fnormals[iFb]
            frame, a = SphericalHarmonics.project_to_frame(self.var[9*iC:9*(iC+1)], stop_threshold=1e-3, nrml_cstr=nrml)
            self.var[9*iC:9*(iC+1)] = a
            self.frames[iC] = frame
            
    def _laplacian(self):
        lap = laplacian_tetrahedra(self.mesh)
        return sp.kron(lap, sp.eye(9), format="csc")

    def _compute_constraints(self):
        self.log(" | Compute constraints matrix")
        nvar = len(self.var)
        ncstr = 0
        rows, cols, coeffs = [], [], []
        cstrRHS = []

        Z = Vec(0.,0.,1.)
        for iC in self._cell_on_bnd:
            if self._cell_on_bnd[iC]>1 :
                # completely lock the frame
                rows += [ncstr + _r for _r in range(9)]
                cols += [9*iC + _c for _c in range(9)]
                coeffs += [1]*9
                cstrRHS += [self.var[9*iC+_c] for _c in range(9)]
                ncstr += 9

        for iFb in self._boundary_mesh.id_faces:
            iF = self.mesh.boundary_connectivity.b2m_face[iFb]
            iC = self.mesh.connectivity.face_to_cells(iF)[0]
            if self._cell_on_bnd[iC] != 1 : 
                continue
            axis = geom.cross(Z, self._fnormals[iFb])
            angle = geom.angle_2vec3D(Z, self._fnormals[iFb])
            if axis.norm()>1e-8: 
                axis = Vec.normalized(axis) * angle
            nrml_frame = SphericalHarmonics.from_vec3(axis)
            rows += [ncstr]*9
            cols += [9*iC + _c for _c in range(9)]
            coeffs += [nrml_frame[_c] for _c in range(9)]
            cstrRHS.append(nrml_frame.dot(self.var[9*iC:9*(iC+1)]))
            ncstr += 1

        cstrMat = sp.csc_matrix((coeffs, (rows, cols)), shape=(ncstr, nvar))
        cstrRHS = np.array(cstrRHS)
        # print(geom.norm(cstrMat.dot(self.var) - cstrRHS))
        return cstrMat, cstrRHS

    def optimize(self):
        self._check_init()
        self.log(" | Compute Laplacian")
        lap = self._laplacian()
        Q = lap #lap.transpose() @ lap
        cstrMat, cstrRHS = self._compute_constraints()
        self.log(" | Initial solve of linear system")
        instance = OSQP()
        instance.setup(Q, None, A=cstrMat, l=cstrRHS, u=cstrRHS, verbose=self.verbose, polish=True, check_termination=10, 
                        adaptive_rho=True,  linsys_solver=utils.get_osqp_lin_solver())
        res = instance.solve()
        self.var = res.x

        if self.n_smooth>0:
            alpha = self.smooth_attach_weight or 1e-3
            self.log(" | Solve linear system {} times with diffusion".format(self.n_smooth))
            Id = sp.eye(self.var.size, format="csc")
            mat = Q + alpha * Id
            instance = OSQP()
            instance.setup(mat,self.var, A=cstrMat, l=cstrRHS, u=cstrRHS)
            for _ in range(self.n_smooth):
                self.normalize()
                instance.update(q = -alpha * self.var)
                res = instance.solve()
                self.var = res.x

        self.log(" | Final Normalize")
        self.normalize()
        self.smoothed = True

    @property
    def singular_vertices(self):
        if self._singul_vertices is None :
            self.flag_singularities()
        return self._singul_vertices

    @property
    def singular_edges(self):
        if self._singul_edges is None:
            self.flag_singularities()
        return self._singul_edges

    @property
    def singularity_graph(self):
        singu_graph = RawMeshData()
        indir = dict()
        for i,v in enumerate(self.singular_vertices):
            indir[v] = i
            singu_graph.vertices.append(self.mesh.vertices[v])
        for e in self.singular_edges:
            A,B = self.mesh.edges[e]
            singu_graph.edges.append((indir[A], indir[B]))
        return PolyLine(singu_graph)

    def flag_singularities(self):
        self._check_init()
        ZERO_THRESHOLD = 0.1
        self.log("Flag Singularities")
        self._singul_vertices = self.mesh.vertices.create_attribute("singuls", bool)
        self._singul_edges = self.mesh.edges.create_attribute("singuls", float)
        for e in self.mesh.id_edges:
            r = Rotation.identity()
            for c1, c2 in utils.cyclic_pairs(self.mesh.connectivity.edge_to_cell(e)):
                r = match_rotation(self.frames[c1], self.frames[c2]) * r
            if r.magnitude()>ZERO_THRESHOLD:
                self._singul_edges[e] = r.magnitude()
        
        for e in self._singul_edges:
            A,B = self.mesh.edges[e]
            self._singul_vertices[A] = True
            self._singul_vertices[B] = True

    def export_frames_to_file(self, file_path :str):
        """
        Exports the frames as 3x3 matrices in a .frame file, with syntax:

            FRAME
            number_of_frames
            a1x a1y a1z b1x b1y b1z c1x c1y c1z
            a2x a2y a2z b2x b2y b2z c2x c2y c2z
            ...
            anx any anz bnx bny bnz cnx cny cnz
            END

        Parameters:
            file_path (str): path to the file
        """
        self.compute_frame_from_sh()
        with open(file_path, 'w') as f:
            f.write("FRAME\n")
            f.write(f"{len(self.mesh.cells)}\n")
            for ic in self.mesh.id_cells:
                mat = self.frames[ic].as_matrix()
                f.write(f"{mat[0][0]} {mat[1][0]} {mat[2][0]} {mat[0][1]} {mat[1][1]} {mat[2][1]} {mat[0][2]} {mat[1][2]} {mat[2][2]}\n")
            f.write("END\n")

    def export_as_mesh(self) -> SurfaceMesh:
        """
        Exports the frame field as a mesh for visualization.
        
        Returns:
            SurfaceMesh: The frame field as a surface mesh object, where small cubes represent frames
        """
        self._check_init()
        L = attributes.mean_edge_length(self.mesh,100)
        FFMesh = RawMeshData()
        col = FFMesh.faces.create_attribute("color", float, 3)
        RED,GREEN,BLUE = Vec(1.,0.,0), Vec(0.,1.,0.), Vec(0.,0.,1.)
        barycenters = attributes.cell_barycenter(self.mesh, persistent=False)

        for iC in self.mesh.id_cells:
            r : Rotation = self.frames[iC]
            cube_v = axis_aligned_cube()
            cube_v = transform.scale(cube_v, 0.3*L, orig=Vec(0.,0.,0.))
            cube_v = transform.rotate(cube_v, r, orig=Vec(0.,0.,0.))
            cube_v = transform.translate(cube_v, barycenters[iC])
            FFMesh.vertices += cube_v.vertices
            FFMesh.faces += [tuple((8*iC+x for x in F)) for F in cube_v.faces]

            col[12*iC] = RED
            col[12*iC+1] = RED
            col[12*iC+10] = RED
            col[12*iC+11] = RED

            col[12*iC+2] = GREEN
            col[12*iC+3] = GREEN
            col[12*iC+6] = GREEN
            col[12*iC+7] = GREEN
            
            col[12*iC+4] = BLUE
            col[12*iC+5] = BLUE
            col[12*iC+8] = BLUE
            col[12*iC+9] = BLUE
        return SurfaceMesh(FFMesh)



from .data_container import DataContainer, CornerDataContainer
from ..geometry import Vec
from .. import utils
from .. import config

class RawMeshData:
    """
    A base container class to store all the data relative to a mesh.
    Serves an intermediate between io parsers or manual inputs and instantiated (and typed) mesh classes.
    """

    def __init__(self, mesh=None):
        self.vertices = DataContainer(id="vertices") if mesh is None else mesh.vertices
        self.edges = DataContainer(id="edges") if (mesh is None or not hasattr(mesh,"edges")) else mesh.edges

        self.faces = DataContainer(id="faces") if (mesh is None or not hasattr(mesh,"faces")) else mesh.faces
        self.face_corners = CornerDataContainer(id="face_corners") if (mesh is None or not hasattr(mesh,"face_corners")) else mesh.face_corners
       
        self.cells = DataContainer(id="cells") if (mesh is None or not hasattr(mesh,"cells")) else mesh.cells
        self.cell_corners = CornerDataContainer(id="cell_corners") if (mesh is None or not hasattr(mesh,"cell_corners")) else mesh.cell_corners
        self.cell_faces = CornerDataContainer(id="cell_faces") if (mesh is None or not hasattr(mesh,"cell_faces")) else mesh.cell_faces

        self._dimensionality : int = None
        self._prepared : bool = False
    
    @property
    def id_vertices(self):
        """
        Shortcut for `range(len(self.vertices))`
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for `range(len(self.edges))`
        """
        return range(len(self.edges))

    @property
    def id_faces(self):
        """
        Shortcut for `range(len(self.faces))`
        """
        return range(len(self.faces))

    @property
    def id_cells(self):
        """
        Shortcut for `range(len(self.cells))`
        """
        return range(len(self.cells))

    @property
    def id_facecorners(self):
        """
        Shortcut for `range(len(self.face_corners))`
        """
        return range(len(self.face_corners))

    @property
    def id_cellcorners(self):
        """
        Shortcut for `range(len(self.cell_corners))`
        """
        return range(len(self.cell_corners))

    @property
    def dimensionality(self) -> int:
        """
        Dimensionality of the manifold represented by the data.

        - If only vertices -> 0

        - If vertices and edges (embedded graph) -> 1

        - If faces (surface manifold) -> 2

        - If cells (volume manifold) -> 3

        Returns:
            int: 0,1,2 or 3
        """
        if self._dimensionality is None:
            self._compute_dimensionality()
        return self._dimensionality

    def _compute_dimensionality(self):
        if not self.cells.empty():
            self._dimensionality = 3
        elif not self.faces.empty():
            self._dimensionality = 2
        elif not self.edges.empty():
            self._dimensionality = 1
        else:
            self._dimensionality = 0

    def prepare(self) -> None:
        """
        Prepares the data to have the correct format. This method is called by the constructors of data structures.
        
        First generates explicitely the set of faces from cells and the set of edges from faces. 
        This behavior can be controlled via the [`config.complete_faces_from_cells`][mouette.config.complete_faces_from_cells] and 
        the [`config.complete_edges_from_faces`][mouette.config.complete_edges_from_faces] global config variables.
        
        Then, treatments are applied on each DataContainer:

        - On vertices : casts 3D vectors to [`mouette.Vec`](mouette.geometry.vector.Vec)
        
        - On edges : sorts edge tuples to satisfy edge convention (smallest index first)
        
        - On faces : nothing

        - On face corners : generates the face_corners container if empty
        
        - On cells : nothing

        - On cell corners : generates the cell_corners container if empty

        """
        if self._prepared : return

        # call completion before preparation
        if config.complete_faces_from_cells:
            self._complete_faces_from_cells()

        # call edge completion ***after*** face completion
        if config.complete_edges_from_faces : 
            self._complete_edges_from_faces() 

        self._prepare_vertices()
        self._prepare_edges()
        self._prepare_faces()
        self._generate_face_corners()
        self._prepare_cells()
        self._generate_cell_corners()
        self._generate_cell_faces()
        self._compute_dimensionality()
        self._prepared = True

    def _prepare_vertices(self):
        for iv in self.id_vertices:
            self.vertices[iv] = Vec(self.vertices[iv])

    def _prepare_edges(self):
        N = len(self.vertices)

        def is_valid(a,b):
            return a!=b and 0<=a<N and 0<=b<N

        # Filter invalid edges (like (x,x)) and make the other immutable
        edges_invalid = any((not is_valid(a,b) for a,b in self.edges))
        if edges_invalid:
            # Rebuild the edge container
            new_edges = DataContainer(id="edges")
            new_attrs = dict()
            old_attrs = dict()
            for attr_name in self.edges.attributes:
                old_attrs[attr_name] = self.edges.get_attribute(attr_name)
                new_attrs[attr_name] = new_edges.create_attribute(attr_name, old_attrs[attr_name].type, old_attrs[attr_name].elemsize)
            n = 0
            for ie in self.id_edges:
                a,b = self.edges[ie]
                if is_valid(a,b):
                    new_edges.append(utils.keyify(a,b))
                    for name in new_attrs:
                        if ie in old_attrs[name]: # keep sparsity of the attribute
                            new_attrs[name][n] = old_attrs[name][ie]
                    n+=1
            self.edges = new_edges
        else:
            for ie in self.id_edges:
                self.edges[ie] = utils.keyify(self.edges[ie])

    def _prepare_faces(self):
        pass

    def _generate_face_corners(self):
        nc = len(self.face_corners)
        nf = sum([len(f) for f in self.faces])
        if nc == 0 or nc!= nf :
            # corners were not or badly generated
            self.face_corners._elem = []
            self.face_corners._adj = []
            for iF,F in enumerate(self.faces):
                for v in F:
                    self.face_corners.append(v,iF)

    def _prepare_cells(self):
        pass

    def _generate_cell_corners(self):
        nce = len(self.cell_corners._elem)
        nca = len(self.cell_corners._adj)
        if nce==0 or nca==0:
            # corners were not or badly generated
            if nca==0 and nce>0:
                # build only adjacency
                self.cell_corners._adj = []
                for iC,C in enumerate(self.cells):
                    self.cell_corners._elem += [iC]*len(C)
            else:
                # build both containers
                self.cell_corners._elem = []
                self.cell_corners._adj = []
                for iC,C in enumerate(self.cells):
                    for v in C:
                        self.cell_corners.append(v,iC)

    def _generate_cell_faces(self):
        nce = len(self.cell_faces._elem)
        nca = len(self.cell_faces._adj)
        if nca==0 or nce==0:
            # cell faces were not generated completely

            face_id = dict() # first invert face indirection
            for iF,F in enumerate(self.faces):
                key = utils.keyify(F)
                face_id[key] = iF

            for iC,C in enumerate(self.cells):
                if len(C)==4:
                    # cell is tetrahedron
                    v0,v1,v2,v3 = C
                    # convention: face fi does not contain vertex vi
                    faces_C = [(v1,v3,v2), (v0,v2,v3), (v3,v1,v0), (v0,v1,v2)]
                elif len(C)==8:
                    # cell is hexahedron : TODO
                    v1,v2,v3,v4,v5,v6,v7,v8 = C
                    faces_C = [
                        (v1,v2,v3,v4),
                        (v5,v6,v7,v8),
                        (v1,v4,v8,v5),
                        (v1,v2,v6,v5),
                        (v2,v3,v7,v6),
                        (v3,v4,v8,v7)
                    ]
                for face in faces_C:
                    self.cell_faces._elem.append(face_id[utils.keyify(face)])
                    if nca!=0: 
                        self.cell_faces._adj.append(iC)

    def _complete_edges_from_faces(self):
        if self.faces.empty() : return # nothing to do
        
        # Create hard edges attribute for already existing edges
        hard_edges = self.edges.create_attribute("hard_edges", bool)
        for e in self.id_edges: hard_edges[e] = True
        
        edge_set = set([utils.keyify(e) for e in self.edges])
        for f in self.faces:
            nf = len(f)
            for i in range(nf):
                edge = utils.keyify(f[i], f[(i+1)%nf])
                if edge not in edge_set:
                    edge_set.add(edge)
                    self.edges.append(edge)

    def _complete_faces_from_cells(self):
        if self.cells.empty() : return # nothing to do
        face_set = set([utils.keyify(f) for f in self.faces])
        for C in self.cells:
            faces_C = []
            if len(C)==8:
                # hexahedron
                v1,v2,v3,v4,v5,v6,v7,v8 = C
                faces_C = [
                    (v1,v2,v3,v4),
                    (v5,v6,v7,v8),
                    (v1,v4,v8,v5),
                    (v1,v2,v6,v5),
                    (v2,v3,v7,v6),
                    (v3,v4,v8,v7)
                ]
            elif len(C)==4:
                # tetrahedron
                v0,v1,v2,v3 = C
                # convention: face fi does not contain vertex vi
                faces_C = [
                    (v1,v3,v2), 
                    (v0,v2,v3), 
                    (v3,v1,v0), 
                    (v0,v1,v2)
                ]
            
            for face in faces_C:
                face_key = utils.keyify(face)
                if face_key not in face_set:
                    face_set.add(face_key)
                    self.faces.append(face)



from .adjacency import adjacency_matrix, vertex_to_edge_operator, vertex_to_face_operator
from .laplacian_op import graph_laplacian, laplacian, laplacian_triangles, laplacian_tetrahedra, volume_laplacian, area_weight_matrix, area_weight_matrix_faces, volume_weight_matrix, cotan_edge_diagonal
from .gradient import gradient


from ..mesh.mesh_data import DataContainer, RawMeshData
from ..mesh.datatypes import *
from ..geometry import Vec
import numpy as np
from ..utils.argument_check import InvalidRangeArgumentError

def de_casteljau(P:list, t:float):
    """De Casteljau's algorithm for evaluating a Bezier curve B_P at some value of the parameter t

    Args:
        P (list): Control points of the Bezier curve
        t (float): parameter value

    Raises:
        InvalidRangeArgumentError: if t is not in the [0;1] interval

    Returns:
        Vec: the position in space of B_P(t) 
    """
    if not 0 <= t <= 1:
        raise InvalidRangeArgumentError("t", t, "in [0,1]")
    coeffs = [x for x in P]
    order = len(P)-1
    for j in range(order):
        for i in range(order - j):
            coeffs[i] = t*coeffs[i+1] + (1-t)*coeffs[i]
    return coeffs[0]


class BezierCurve:

    def __init__(self, control_points):
        self.pts = DataContainer([Vec(x) for x in control_points], id="control_points")

    @property
    def order(self):
        return len(self.pts)-1
    
    def evaluate(self, t):
        return de_casteljau(self.pts, t)
    
    def as_polyline(self, n_pts:int=100, custom_pos=None):
        if custom_pos is not None:
            points = custom_pos
        else:
            points = np.linspace(0,1,n_pts)
        out = RawMeshData()
        uvs = out.vertices.create_attribute("t",float)
        for it,t in enumerate(points):
            pt_t = self.evaluate(t)
            if pt_t.size == 2:
                out.vertices.append(Vec(pt_t[0], pt_t[1], 0.))
            elif pt_t.size == 3:
                out.vertices.append(pt_t)
            uvs[it] = t
        for i in range(n_pts-1):
            out.edges.append((i,i+1))
        return PolyLine(out)
        

class BezierPatch:
    def __init__(self, control_points):
        self.pts = [[Vec(x) for x in l] for l in control_points]

    @property
    def order(self) -> int:
        return (len(self.pts)-1, len(self.pts[0])-1)
    
    def _evaluate_row(self,u):
        return [de_casteljau(self.pts[i],u) for i in range(len(self.pts))]

    def evaluate(self,u,v) -> Vec:
        return de_casteljau(self._evaluate_row(u),v)

    def as_surface(self,n1=20,n2=20) -> SurfaceMesh:
        out = RawMeshData()
        U = np.linspace(0,1,n1)
        V = np.linspace(0,1,n2)
        uvs = out.vertices.create_attribute("uv_coords", float, 2)
        k = 0

        # evaluate vertices
        for i in range(n1):
            q = self._evaluate_row(U[i])
            for j in range(n2):
                out.vertices.append(de_casteljau(q,V[j]))
                uvs[k] = Vec(U[i],V[j])
                k+=1
                
        # add faces
        for i in range(n1-1):
            for j in range(n2-1):
                out.faces.append(( i*n1+j, i*n1+j+1, (i+1)*n1+j+1, (i+1)*n1+j))
        return SurfaceMesh(out)



from ..mesh.datatypes import *
from ..mesh.mesh_attributes import ArrayAttribute
from .. import attributes
from .features import FeatureEdgeDetector
from ..geometry import Vec
from .. import geometry as geom
from .. import utils

import numpy as np
import math
from abc import ABC, abstractmethod

class SurfaceConnection(ABC):
    """
    Abstract base class for surface connections.
    """
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, feat : FeatureEdgeDetector = None ):
        self.mesh = mesh
        self.feat = feat or FeatureEdgeDetector(only_border=True, compute_feature_graph=False, verbose=False)(self.mesh)
        self._baseX : np.ndarray = None
        self._baseY : np.ndarray = None
        self._transport : dict = None
        self._initialize()
    
    @abstractmethod
    def _initialize(self):
        pass

    def transport(self, iA : int, iB : int) -> float:
        """_summary_

        Args:
            iA (int): first vertex index
            iB (int): second vertex index

        Returns:
            float: angle needed to change from basis of iA to basis of iB
        """
        return self._transport[(iA,iB)]

    def base(self, i:int):
        """Local basis of element i

        Args:
            i (int): index

        Returns:
            (Vec,Vec): the X vector and Y vector of the tangent basis.

        Note:
            `cross(X,Y)` is the normal of element i
        """
        return self._baseX[i], self._baseY[i]

    def project(self, V:Vec, i:int):
        return Vec(self._baseX[i].dot(V), self._baseY[i].dot(V))


class SurfaceConnectionVertices(SurfaceConnection):
    """
    Local bases and parallel transport defined between tangent planes of the mesh at the vertices

    Reference:
        _Globally Optimal Direction Fields_, Knöppel et al. (2013)
    
    Args:
        mesh (SurfaceMesh): the supporting mesh
        feat (FeatureEdgeDetector, optional): feature edges of the mesh. If not provided, will be initialized with a FeatureEdgeDetector object that flags only the boundary. Defaults to None.
    
    Keyword Args:
        vnormals (Attribute): An attribute representing the vertex normals. If not provided, will be computed at initialization
        angles (Attribute) : An attribute representing angles at corners of faces. If not provided, will be computed at initialization

    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh,  feat : FeatureEdgeDetector = None, **kwargs):
        self.vnormals = kwargs.get("vnormals", None)
        if self.vnormals is None: 
            if mesh.vertices.has_attribute("normals"):
                self.vnormals = mesh.vertices.get_attribute("normals")
            else :
                self.vnormals = attributes.vertex_normals(mesh)
        
        self.angles = kwargs.get("angles", None)
        if self.angles is None:
            if mesh.face_corners.has_attribute("angles"):
                self.angles = mesh.face_corners.get_attribute("angles")
            else:
                self.angles = attributes.corner_angles(mesh)
        
        self.total_angle = ArrayAttribute(float, len(mesh.vertices))
        for iC, C in enumerate(mesh.face_corners):
            self.total_angle[C] = self.total_angle[C] + self.angles[iC] 
        super().__init__(mesh, feat)

    def _initialize(self):
        n_vert = len(self.mesh.vertices)
        self._baseX = ArrayAttribute(float, n_vert, 3)
        self._baseY = ArrayAttribute(float, n_vert, 3)
        self._transport = dict()

        for u in range(n_vert):
            # find basis vector -> first edge
            P = Vec(self.mesh.vertices[u])
            N = self.vnormals[u]
            # extract basis edge
            vert_u = self.mesh.connectivity.vertex_to_vertices(u)[::-1]
            E = self.mesh.vertices[vert_u[0]] - P
            X = Vec.normalized(E - np.dot(E,N)*N) # project on tangent plane
            self._baseX[u] = X
            self._baseY[u] = geom.cross(N,X)

            # initialize angles of every edge in this basis
            ang = 0.
            if u in self.feat.feature_vertices:
                dfct = self.feat.corners[u] * 2 * np.pi / self.feat.corner_order # target defect (multiple of pi/order)
                for v in vert_u:
                    T = self.mesh.connectivity.direct_face(u,v)
                    self._transport[(u,v)] = ang * dfct / self.total_angle[u]
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    if c is None: continue
                    ang += self.angles[c]
            else:
                # normal vertex in interior -> flatten to 2pi
                for v in vert_u:
                    T = self.mesh.connectivity.direct_face(u,v)
                    self._transport[(u,v)] = ang * 2 * np.pi / self.total_angle[u]
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    ang += self.angles[c]

class FlatConnectionVertices(SurfaceConnection):
    """
    Surface connection on vertices where all local bases are taken as the canonical basis ([1 0 0], [0 1 0]).
    /!\\ This only makes sense if the considered surface is embedded in R^2, meaning that there is no curvature.

    Args:
        mesh (SurfaceMesh): the supporting mesh
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh):
        super().__init__(mesh, None)
        self._baseX = Vec.X()
        _,_,sample_normal = geom.face_basis(*(self.mesh.vertices[_u] for _u in self.mesh.faces[0][:3]))
        if sample_normal.z<0:
            self._baseY = -Vec.Y()
        else:
            self._baseY = Vec.Y()

    def _initialize(self):
        return # nothing to do

    def transport(self, iA,iB):
        E = self.mesh.vertices[iB] - self.mesh.vertices[iA]
        return np.arctan2(E.y,E.x)

    def base(self, i:int):
        # Same base everywhere
        return self._baseX, self._baseY

    def project(self, V:Vec, i:int):
        # projection does not depend on the considered vertex i
        return Vec(self._baseX.dot(V), self._baseY.dot(V))

class SurfaceConnectionFaces(SurfaceConnection):
    """
    Local bases and parallel transport defined between tangent planes of the mesh at the vertices

    Reference:
        _Globally Optimal Direction Fields_, Knöppel et al. (2013)

    Args:
        mesh (SurfaceMesh): the supporting mesh
        feat (FeatureEdgeDetector, optional): feature edges of the mesh. If not provided, will be initialized with a FeatureEdgeDetector object that flags only the boundary. Defaults to None.
    """
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, feat : FeatureEdgeDetector = None) -> None:
        super().__init__(mesh, feat)

    def _initialize(self):
        #### Initialize bases
        NF = len(self.mesh.faces)
        self._baseX, self._baseY = ArrayAttribute(float, NF, 3), ArrayAttribute(float, NF, 3)
        self._transport = dict()
        for id_face, (A,B,C) in enumerate(self.mesh.faces):
            feat = [self.mesh.connectivity.edge_id(_u,_v) in self.feat.feature_edges for (_u,_v) in [(A,B), (B,C), (C,A)]]
            if np.any(feat):
                # face has feature edge -> feature should be the X coordinate of the basis
                A,B,C = utils.offset([A,B,C], np.argmax(feat))
            pA,pB,pC = (self.mesh.vertices[_v] for _v in (A,B,C))
            X,Y,_ = geom.face_basis(pA,pB,pC) # local basis of the triangle (ignore normal)
            self._baseX[id_face] = X
            self._baseY[id_face] = Y

        #### Initialize Parallel Transport
        for e in self.mesh.interior_edges:
            A,B = self.mesh.edges[e]
            pA,pB = self.mesh.vertices[A], self.mesh.vertices[B]
            E = geom.Vec(pB-pA)
            T1,T2 = self.mesh.connectivity.edge_to_faces(A,B)
            X1,Y1 = self._baseX[T1], self._baseY[T1]
            X2,Y2 = self._baseX[T2], self._baseY[T2]
            angle1 = math.atan2( geom.dot(E,Y1), geom.dot(E,X1))
            angle2 = math.atan2( geom.dot(E,Y2), geom.dot(E,X2))
            self._transport[(T1,T2)] = angle1 - angle2
            self._transport[(T2,T1)] = angle2 - angle1

class FlatConnectionFaces(SurfaceConnection):
    """
    Surface connection on faces where all local bases are taken as the canonical basis ([1 0 0], [0 1 0]).
    /!\\ This only makes sense if the considered surface is embedded in R^2, meaning that there is no curvature.

    Args:
        mesh (SurfaceMesh): the supporting mesh
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh):
        super().__init__(mesh, None)
        self._baseX = Vec.X()
        _,_,sample_normal = geom.face_basis(*(self.mesh.vertices[_u] for _u in self.mesh.faces[0][:3]))
        if sample_normal.z<0:
            self._baseY = -Vec.Y()
        else:
            self._baseY = Vec.Y()

    def _initialize(self):
        return # nothing to do

    def transport(self, iA,iB):
        return 0.

    def base(self, i:int):
        # Same base everywhere
        return self._baseX, self._baseY

    def project(self, V:Vec, i:int):
        # projection does not depend on the considered vertex i
        return Vec(self._baseX.dot(V), self._baseY.dot(V))


from ..mesh_data import RawMeshData
from ...geometry import Vec
from collections import deque

def import_off(path : str):
    """
    Read a .off file : 
		/////////////////
        OFF
        #v #f #e
        v1_x v1_y v1_z
        v2_x v2_y v2_z
        ...

        vn_x vn_y vn_z
        3 v1_1 v1_2 v1_3
        3 v2_1 v2_2 v2_3
        ...
        3 vm_1 vm_2 vm_3
        /////////////////

        if surfacic, 3 is given at beginning of each simplex line (triangle), otherwise 4 (tetrahedron)
    Parameters:
        path (str): the input file path
    """

    with open(path, 'r' ) as file:
        output = parse_off_data(file.readlines())
    return output

def parse_off_data(data):
    output = RawMeshData()
    
    data = [x.strip().split() for x in data]
    # remove empty lines from data
    data = deque([x for x in data if x])

    header = data.popleft()[0]
    if (header != "OFF"): # file always starts with OFF
        raise Exception("Import OFF file : OFF header missing.")

    nv,nf,ne = (int(u) for u in data.popleft())

    for _ in range(nv):
        vertex = [float(u) for u in data.popleft()]
        output.vertices.append(Vec(vertex))

    i_f, i_c = 0, 0
    for _ in range(nf):
        simplex = data.popleft()
        nvi = int(simplex[0])
        if nvi==3:
            face = [int(u) for u in simplex[1:nvi+1]]
            output.faces.append(face)
            output.face_corners += [(x,i_f) for x in face]
            i_f += 1
        elif nvi==2:
            a,b = simplex[1], simplex[2]
            output.edges.append((min(a,b), max(a,b)))
        elif nvi==4:
            cell = [int(u) for u in simplex[1:nvi+1]]
            output.cells.append(cell)
            output.cell_corners += [(x,i_c) for x in cell]
            i_c += 1
    return output

def export_off(mesh, path):
    with open( path, 'w' ) as ofile:
        ofile.write("OFF\n")
        nums = "{} {} {}\n".format(len(mesh.vertices), len(mesh.faces), len(mesh.edges))
        ofile.write(nums)
        for vtx in mesh.vertices:
            ofile.write(' '.join(['{}'.format(v) for v in vtx])+'\n')
        for face in mesh.faces:
            str_face = "{} ".format(len(face))
            str_face += " ".join([str(v) for v in face])
            ofile.write(str_face + "\n")


from ..mesh_data import RawMeshData
from .base import Mesh

class PointCloud(Mesh):
    """
    A data structure for representing point clouds

    Attributes:
        vertices (DataContainer): the container for all vertices

        __str__: Representation of the object and its elements as a string.
    """

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 0, data)

    def __str__(self) -> str:
        out = "mouette.mesh.PointCloud object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        return out

    def append(self,x):
        """
        Shortcut for `self.vertices.append(x)`, since we can only append elements in the 'vertices' container
        """
        self.vertices.append(x)

    @property
    def id_vertices(self):
        """
        Shortcut for `range(len(self.vertices))`
        """
        return range(len(self.vertices))


from ..mesh.datatypes import *
from .. import geometry as geom
import scipy.sparse as sp
import numpy as np

@forbidden_mesh_types(PointCloud)
def adjacency_matrix(mesh : Mesh, weights="one") -> sp.coo_matrix:
    """
    Computes and returns the adjacency matrix of a mesh, that is a (sparse) matrix M such that
        
        M[i,j] = M[j,i] = weights[i,j] if (i,j) is an edge of the mesh
        
        M[i,j] = 0 otherwise


    Args:
        mesh (Mesh): the input mesh
        weights (str, optional): How to weight the edges of the matrix. Options are:
            - "one" : every edge is 1
            - "length" : every edge has weight corresponding to its length
            - a custom dict edge_id:int -> weight:float
            Defaults to "one".

    Returns:
        scipy.sparse.coo_matrix
    """
    if not ( (isinstance(weights, str) and weights in ("one", "length")) or isinstance(weights, dict)):
        raise Exception("weights should be 'ones', 'length' or a custom dict")

    n = len(mesh.vertices)
    m = len(mesh.edges)

    # build matrix coefficients
    if weights == "one":
        vals = np.ones(2*m)
    elif weights == "length":
        vals = np.zeros(2*m)
        for e,(a,b) in enumerate(mesh.edges):
            d = geom.distance(mesh.vertices[a], mesh.vertices[b])
            vals[2*e] = d
            vals[2*e+1] = d
    else:
        vals = np.zeros(2*m)
        for e in mesh.id_edges:
            vals[2*e] = weights[e]
            vals[2*e+1] = weights[e]
    
    # build row and cols indices
    rows, cols = np.zeros(2*m), np.zeros(2*m)
    for e,(a,b) in enumerate(mesh.edges):
        rows[2*e] = a
        rows[2*e+1] = b
        cols[2*e] = b
        cols[2*e+1] = a

    return sp.coo_matrix( (vals, (rows,cols)), shape=(n,n))

@forbidden_mesh_types(PointCloud)
def vertex_to_edge_operator(mesh : Mesh, oriented:bool = False) -> sp.csc_matrix:
    """Vertices to edges operator. Matrix M of size |V|x|E| where:
        
        M[v,e] = 1 if and only if v is one extremity of edge e.

        if oriented is set, M[v,e] = -1 if v is the origin of the edge and 1 if it is the arrival.

    Args:
        mesh (Mesh): the input mesh
        oriented (bool): whether to consider oriented edge and signed values or not. Defaults to False.

    Returns:
        scipy.sparse.csc_matrix
    """
    n,m = len(mesh.vertices), len(mesh.edges)
    mat = sp.lil_matrix((n,m))
    orig_coeff = -1 if oriented else 1
    for e,(A,B) in enumerate(mesh.edges):
        mat[A,e] = orig_coeff
        mat[B,e] = 1
    return mat.tocsc()

@allowed_mesh_types(SurfaceMesh)
def vertex_to_face_operator(mesh : SurfaceMesh) -> sp.csc_matrix:
    """
    Vertices to face operator. Matrix M of size |V| x |F| where:

        M[v,f] = 1/len(f) if and only if v is one of the vertices of f

    Args:
        mesh (SurfaceMesh): input surface mesh

    Returns:
        scipy.sparse.csc_matrix
    
    See also:
        mouette.attributes.interpolate_vertices_to_faces
    """
    n,m = len(mesh.vertices), len(mesh.faces)
    mat = sp.lil_matrix((m,n))
    for iT,T in enumerate(mesh.faces):
        aT = 1/len(T)
        for V in T:
            mat[iT,V] = aT
    return mat.tocsc()


import numpy as np
from osqp import OSQP
from dataclasses import dataclass
import scipy.sparse as sp
from scipy.optimize import line_search

from tqdm import tqdm, trange
from tqdm.utils import _term_move_up
prefix = _term_move_up() + '\r'

from ..utils import Logger, get_osqp_lin_solver
from .. import geometry as geom

@dataclass
class NewtonParameters:
    """
    Hyper parameters for Newton's algorithm
    """
    N_ITER_MAX    : int   = 1000 # stopping criterion on maximal number of iterations
    MIN_STEP_NORM : float = 1e-6 # stopping criterion, on ||x_n - x_{n-1}||
    MIN_DELTA_F   : float = 1e-8 # stopping criterion on ||f_n - f_{n-1}||
    MIN_GRAD_NORM : float = 1e-8 # stopping criterion on gradient projected on constraints

    alpha_init : float = 1. # starting step size in linesearch
    c1 : float = 1e-4 # Weak Wolfe 1st condition parameter
    rho : float = 0.5 # shrinkage parameter in linesearch

@dataclass
class NewtonVerboseOptions:
    logger_verbose : bool = True  # main verbose for mouette logging
    solver_verbose : bool = False # prints outputs of OSQP solver
    use_tqdm       : bool = False  # progression bar using tqdm
    log_frequency  : int  = 1     # display an info message every 'log_frequency' iterations

class Newton(Logger):
    """
    Newton's algorithm to find the minimum of a function knowing its gradient and hessian.

    References:
        [1] https://won-j.github.io/M1399_000200-2021fall/lectures/22-newton/newton_constr.html
    """

    def __init__(self,
        HP : NewtonParameters = NewtonParameters(), 
        verbose : NewtonVerboseOptions = NewtonVerboseOptions(),
        **kwargs):
        super().__init__("Newton",verbose.logger_verbose)
        self.verbose_options = verbose
        self.HP = HP

        self.X = None # variables

        ### Linear Constraints
        self._cstM = None
        self._cstL = None
        self._cstU = None

        ### Function to optimize
        self._func = None
        self._func_no_grad = None

        ### Others
        self._stop_criterion_instance = None # OSQP instance to compute projected gradient norm

        ### Additionnal parameters
        self._linsys_solver = kwargs.get("lin_solver", get_osqp_lin_solver())

    def register_constraints(self, A : sp.spmatrix, l : np.ndarray = None, u : np.ndarray = None):
        """
        Adds linear constraints to the optimization problem :

        min_X  F(X) s.t.  l <= AX <= u

        Args:
            cstMat (sp.spmatrix): constraint matrix, either dense or sparse
            cstL (np.ndarray, optional): vector l. Defaults to None.
            cstR (np.ndarray, optional): vector u. Defaults to None.
        """

        if isinstance(A, sp.spmatrix):
            self._cstM = A.tocsc()
        elif isinstance(A, np.ndarray):
            assert len(A.shape)==2 # we make sure A is indeed a matrix
            self._cstM = sp.csc_matrix(A)
        self._cstL = l
        self._cstU = u

    def register_function(self, fun, fun_noG=None):
        """
        Adds a function to minimize.

        Args:
            fun (python callable): A function taking a single argument X (np.array) returning a real value F, a gradient G and a hessian matrix H

            fun_noG (python callable, optional): The same function that avoids the computation of the gradient and the hessian for speed purposes. If not provided, will use the function provided above and ignore the other arguments.
        """
        self._func = fun
        if fun_noG is None:
            fun_noG = lambda x : fun(x)[0]
        self._func_no_grad = fun_noG

    def energy(self, X=None, derivatives=False):
        if X is None: X = self.X
        if derivatives:
            return self._func_no_grad(X)
        return self._func(X)

    def optimize(self, x_init: np.ndarray) -> float:
        """
        Alias of the 'run' method.

        Args:
            x_init (np.ndarray): initial values for the variables.

        Returns:
            float: final value of the energy
        """
        return self.run(x_init)

    def run(self, x_init : np.ndarray):
        """
        Runs the optimizer

        Args:
            x_init (np.ndarray): initial values for the variables.

        Returns:
            float: final value of the energy
        """

        self.X = self._find_closest_feasible_solution(x_init) # variable vector
        
        if self.HP.N_ITER_MAX <= 0 :
            # return current energy
            return self.energy(x_init, derivatives=False)

        if self._func is None: 
            # No functions to optimize
            self.log("No function to optimize. Register a function before calling the optimization.")
            return 0.

        # iterable for optimization steps
        if self.verbose_options.logger_verbose and self.verbose_options.use_tqdm:
            print()
            iterobj = trange(self.HP.N_ITER_MAX, total=self.HP.N_ITER_MAX, position=1, leave=False, ncols=100, unit="")
        else:
            iterobj = range(self.HP.N_ITER_MAX)

        try:
            # Iteration loop
            for it in iterobj:
                F,G,H = self.energy()                
                if np.isnan(F):
                    return self._end_optimization(F, "NaN value in energy")
                grad_stop, grad_norm = self._stop_criterion_projgrad(G)
                if grad_stop:
                    return self._end_optimization(F, "projected grad norm on constraints < min_grad_norm")

                osqp_instance = OSQP()
                if self._cstM is not None:
                    # OSQP solve inside iteration loop computes the increment x_{n-1} - x_n, so constraints have to be shifted
                    xCst = self._cstM @ self.X
                    cstL = self._cstL - xCst
                    cstU = self._cstU - xCst
                else:
                    cstL,cstU = None, None

                ### Compute direction of descent
                osqp_instance.setup(H, q=G, A=self._cstM, l=cstL, u=cstU,
                    verbose=self.verbose_options.solver_verbose,
                    eps_abs=1e-3, eps_rel=1e-3,
                    max_iter=100, polish=True, check_termination=10, 
                    adaptive_rho=True, linsys_solver= self._linsys_solver)
                s = osqp_instance.solve().x # computed increment
                if s[0] is None: print(s)

                ### Perform Linesearch
                # ln = line_search(self._func_no_grad, lambda x : self._func(x)[1], self.X, -G)
                # alpha, new_F = ln[0], ln[3]
                # success = (alpha is not None)
                
                # alpha,new_F,success = self._backtracking_linesearch(F,G,s)
                
                alpha, new_F, success = 1., self._func_no_grad(self.X+s), True

                if not success:
                    return self._end_optimization(F, "Linesearch failed")
                self.X += s*alpha
                RelDeltaF = abs(new_F-F)/abs(F)
                if RelDeltaF < self.HP.MIN_DELTA_F:
                    return self._end_optimization(F, "Relative progression of F < ΔF_min")
                step_norm = geom.norm(s*alpha)
                if step_norm < self.HP.MIN_STEP_NORM :
                    return self._end_optimization(F, "Step norm < min_step_norm")

                if self.verbose_options.logger_verbose and self.verbose_options.log_frequency >0 and it%self.verbose_options.log_frequency==0:
                    if self.verbose_options.use_tqdm:
                        tqdm_log = prefix
                        tqdm_log += "F: {:.2E} | Grad: {:.2E} | ".format(F, grad_norm)
                        tqdm_log += "ΔE {:.2E} | Step {:.2E}".format(RelDeltaF, step_norm)
                        tqdm_log += " " * 10
                        tqdm.write(prefix)
                        tqdm.write(tqdm_log)
                    else:
                        log = f"{it+1}/{self.HP.N_ITER_MAX} | "
                        log += "F: {:.2E} | Grad: {:.2E} | ".format(F, grad_norm)
                        log += "ΔE {:.2E} | Step {:.2E}".format(RelDeltaF, step_norm)
                        print(log)

        except KeyboardInterrupt:
            self.log("Manual interruption")
        return self._end_optimization(F, "max iteration reached")

    def _find_closest_feasible_solution(self, x_init):
        if self._cstM is None: 
            # problem is unconstrained, solution is feasible
            return np.copy(x_init)
        nvar = x_init.size
        if self._cstM.shape[1] != nvar:
            raise Exception(f"Constraint matrix size {self._cstM.shape} and number of variables {nvar} mismatch.")
        instance = OSQP()
        instance.setup(sp.eye(nvar, format="csc"), q=-x_init, 
                    A=self._cstM, l=self._cstL, u=self._cstU, 
                    verbose=True, linsys_solver= self._linsys_solver)
        x_proj = instance.solve().x
        return x_proj

    def _backtracking_linesearch(self, F, G, s):
        a = self.HP.alpha_init
        f_step = self._func_no_grad(self.X+a*s)
        Gs = np.dot(G,s)
        n_iter = 0
        n_iter_max = 100
        while f_step >= F + self.HP.c1 * a * Gs and n_iter<n_iter_max:
            a = self.HP.rho * a
            f_step = self._func_no_grad(self.X + a*s)
            n_iter += 1
        return a, f_step, (n_iter < n_iter_max)

    def _stop_criterion_projgrad(self, G):
        # Projected gradient on the orthogonal of space spanned by constraints should have small norm
        if self._cstM is None:
            xnorm = np.sqrt(np.dot(G,G))
        else:
            n = self._cstM.shape[1]
            if self._stop_criterion_instance is None:
                self._stop_criterion_instance = OSQP()
                self._stop_criterion_instance.setup(
                    sp.eye(n,format=("csc")), q=-2*G, A = self._cstM, l = self._cstL, u = self._cstU,
                    verbose=self.verbose_options.solver_verbose)
            else:
                self._stop_criterion_instance.update(q=-2*G)
            x = self._stop_criterion_instance.solve().x
            if x is None: return False, 0.
            xnorm = np.sqrt(np.dot(x,x))
        return xnorm<self.HP.MIN_GRAD_NORM, xnorm

    def _end_optimization(self, energy, message):
        if self.verbose:
            if self.verbose_options.use_tqdm: print("\n")
            self.log("End of optimization: "+message)
            self.log(f"Final Energy: {energy:.4E}")
            if self.verbose_options.use_tqdm: print("\r")
        return energy        


from .vertex2d import _BaseFrameField2DVertices
from ...mesh.mesh_attributes import Attribute
from ...mesh.datatypes import *
from ...import attributes
from ... import geometry as geom
from ...utils.maths import *
from ... import operators

import numpy as np
import scipy.sparse as sp
from scipy.sparse import linalg
from numba import jit, prange

@jit(parallel=True, nopython=True)
def aggregate_mats(curvV, curvE, areas, iV,iE, n):
    nvert = curvV.shape[0]
    total_area = np.zeros(nvert, dtype=np.float64)
    for i in prange(n):
        v,e = iV[i], iE[i]
        curvV[v,:,:] += areas[e]*curvE[e,:,:]
        total_area[v] += areas[e]
    for v in prange(nvert):
        curvV[v,:,:] /= total_area[v]
    return curvV

# version without numba
# def aggregate_mats(curvV, curvE, areas, iV,iE, n):
#     nvert = curvV.shape[0]
#     total_area = np.zeros(nvert, dtype=np.float64)
#     for i in range(n):
#         v,e = iV[i], iE[i]
#         curvV[v,:,:] += areas[e]*curvE[e,:,:]
#         total_area[v] += areas[e]
#     for v in range(nvert):
#         curvV[v,:,:] /= total_area[v]
#     return curvV

class PrincipalDirectionsVertices(_BaseFrameField2DVertices):
    """
    Principal curvature direction estimation using a frame field on vertices.

    Implementation based on 'Restricted Delaunay Triangulations and Normal Cycle',  David Cohen-Steiner and Jean-Marie Morvan, 2003
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        mesh : SurfaceMesh, 
        feature_edges : bool = False,
        verbose : bool = True,
        **kwargs):
        """
        Parameters:
            mesh (SurfaceMesh): the surface mesh on which to perform the estimation
            feature_edges (bool, optional): Flag to take into account feature edges. Defaults to False.
            patch_size (int, optional): Neighborhood size in curvature estimation. Greater means more averaging but also more computation. Defaults to 3.
            curv_threshold (float, optional): Minimal mean curvature value for the directions to be computed. Defaults to 0.01.
            verbose (bool, optional): verbose mode. Defaults to True.
        
        Additionnal Parameters:
            TODO
        
        Note:
            Order of the frame field is fixed at 4 since principal curvature directions form an orthonormal basis.
        """
        
        super().__init__(mesh, 
            4, # order is always 4 
            feature_edges,
            verbose,
            **kwargs
        )
        self.complete_ff : bool = kwargs.get("complete_ff", True)
        self.patch_size : int = kwargs.get("patch_size", 3)
        self.curv_threshold : float = kwargs.get("curv_threshold", 0.01)
        self.face_areas : Attribute = None
        self.curv_mat_vert : np.ndarray = None # shape (|V|,3,3)
        self.M : sp.lil_matrix = None

    def run(self):
        self.log("Compute curvature matrices")
        self.initialize()
        self.log("Optimize")
        self.optimize()
        self.log("Done.")
        return self

    def _initialize_attributes(self):
        super()._initialize_attributes()
        self.face_areas = attributes.face_area(self.mesh)

    def _build_patch_connectivity_matrix(self):
        self.log("Compute vertex neighborhoods")
        V2E = operators.vertex_to_edge_operator(self.mesh)
        self.M = operators.adjacency_matrix(self.mesh).tolil()
        self.M.setdiag(1) # adjacency + Id
        self.M = pow(self.M, self.patch_size-1) @ V2E
        self.M.minimum(1)

    def _initialize_curv_matrices(self):
        self.log("Aggregate curvature matrices on patches")
        curv_mat_edges = attributes.curvature_matrices(self.mesh)
        self.curv_mat_vert = np.zeros((len(self.mesh.vertices),3,3), dtype=np.float64)
        row,col = self.M.nonzero()
        areas = np.zeros(len(self.mesh.edges), dtype=np.float64)
        for e,(a,b) in enumerate(self.mesh.edges):
            areas[e] = sum([self.face_areas[_T] for _T in self.mesh.connectivity.edge_to_faces(a,b) if _T is not None])/2
        self.curv_mat_vert = aggregate_mats(self.curv_mat_vert, curv_mat_edges, areas, row, col, self.M.count_nonzero())

    def initialize(self):
        self._initialize_attributes()
        self._build_patch_connectivity_matrix()
        self._initialize_curv_matrices()
        self._initialize_variables()
        self.initialized = True

    def optimize(self):
        curvW = self.mesh.vertices.create_attribute("curvCoeff", float, dense=True)

        self.log("Computing curvature via SVD")
        for v in self.mesh.id_vertices:
            if v in self.feat.feature_vertices: continue # do not override frames on boundary
            
            U,S,V = np.linalg.svd(self.curv_mat_vert[v,:,:], hermitian=True)
            # three vectors -> normal and two principal components
            # eigenvalue of normal is 0
            # PC are orthogonal (eigenvects of symmetric matrix) -> we rely on the eigenvect of greatest eigenvalue and take orthog direction

            X,Y = self.conn.base(v)
            curv_coeff_v = geom.norm(S)
            if self.n_smooth > 0: curvW[v] = curv_coeff_v

            if curv_coeff_v<self.curv_threshold : # zero matrix -> no curvature information
                self.var[v] = 0 + 0j
            else:
                # representation vector
                eig = V[0,:]
                c = complex(X.dot(eig), Y.dot(eig))
                self.var[v] = (c/abs(c))**4
        
        lap = None
        if self.complete_ff and self.curv_threshold>0:
            lap = operators.laplacian(self.mesh, cotan=self.use_cotan, connection=self.conn, order=4)
            self.log("Completing curvature on flat regions")
            freeAttr = self.mesh.vertices.create_attribute("free", bool)
            A = operators.area_weight_matrix(self.mesh)
            fixedInds,freeInds = [],[]
            for v in self.mesh.id_vertices:
                if abs(self.var[v])>1e-8:
                    fixedInds.append(v)
                else:
                    freeAttr[v] = True
                    freeInds.append(v)
            if freeInds:
                lapI = lap[freeInds,:][:,freeInds]
                lapB = lap[freeInds,:][:,fixedInds]
                # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
                valB = lapB.dot(self.var[fixedInds]) # right hand side
                res = linalg.spsolve(lapI, - valB)
                self.var[freeInds] = res

        if self.n_smooth > 0:
            self.log(f"Smoothing frame field {self.n_smooth} times with diffusion.")
            # diffuse the curvature results to get a smoother results (especially where curvature was not defined)
            if lap is None:
                lap = operators.laplacian(self.mesh, cotan=self.use_cotan, connection=self.conn, order=4)
                A = operators.area_weight_matrix(self.mesh)
            alpha = self.smooth_attach_weight or 1.
            self.log("Attach Weight", alpha)

            if len(self.feat.feature_vertices)>0:
                # Build fixed/free vertex partition
                fixedInds,freeInds = [],[]
                for v in self.mesh.id_vertices:
                    if v in self.feat.feature_vertices:
                        fixedInds.append(v)
                    else:
                        freeInds.append(v)
                lapI = lap[freeInds,:][:,freeInds]
                lapB = lap[freeInds,:][:,fixedInds]
                AI = A[freeInds, :][:,freeInds]
                # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
                valB = lapB.dot(self.var[fixedInds]) # right hand side
                mat = lapI -  alpha * AI
                for _ in range(self.n_smooth):
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, - valB - valI2)
                    self.var[freeInds] = res
                    self.normalize()
            else:
                mat = lap - alpha * A
                for _ in range(self.n_smooth):
                    valI2 = alpha * A.dot(self.var)
                    self.var = linalg.spsolve(mat, - valI2)
                    self.normalize()
        else: 
            self.normalize()
        self.smoothed = True


import setuptools

with open("README.md", "r") as file_handle:
    long_description = file_handle.read()

setuptools.setup(
    name="mouette",
    version='1.2.5',
    author="GCoiffier",
    description="Mesh, Tools and Geometry Processing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GCoiffier/mouette",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=[
        "aenum",
        'numpy',
        'scipy',
        'osqp',
        'numba',
        'tqdm',
        'stl-reader',
        'pyminiply',
    ]
)



from ..mesh.mesh_data import RawMeshData
from ..utils import keyify
from ..mesh.datatypes import *
from ..mesh.mesh_attributes import Attribute

@allowed_mesh_types(SurfaceMesh)
def extract_border_cycle(mesh : SurfaceMesh, starting_point : int = None):
    """
    Extracts a list of vertices that are on the border of the mesh.

    Parameters:
        mesh (Mesh): The mesh
        starting_point (int): Origin point. Should be on border
    
    Returns:
        (list): list of vertices on border in order
        (list): list of edges on border in order

    Raises:
        Exception: Fails if 'starting_point' not on the border
    """
    if len(mesh.boundary_vertices)==0 : return []
    if starting_point is None:
        starting_point = mesh.boundary_vertices[0]
    if not mesh.is_vertex_on_border(starting_point) : 
        raise Exception("Starting point (vertex {}) is not on mesh border".format(starting_point))

    vborder, eborder = [starting_point], []
    point1, point2 = starting_point, mesh.connectivity.vertex_to_vertices(starting_point)[0]
    nvisited = 0
    MAX_VISITED = len(mesh.vertices)
        
    while point2 != starting_point and nvisited < MAX_VISITED:
        # while we have not come back to origin
        vborder.append(point2)
        eborder.append(mesh.connectivity.edge_id(point1, point2))
        for v in mesh.connectivity.vertex_to_vertices(point2):
            if mesh.is_vertex_on_border(v) and v!=point1:
                point1, point2 = point2, v
                break
        nvisited += 1
    eborder.append(mesh.connectivity.edge_id(point1, point2)) # add last edge to close the cycle
    return vborder, eborder

@allowed_mesh_types(SurfaceMesh)
def extract_border_cycle_all(mesh : SurfaceMesh) -> list:
    """
    Extracts all the border cycles of a mesh and returns a list of list of vertices

    Parameters:
        mesh (Mesh): The mesh
    
    Returns:
        (list): list of list of vertices
    """
    visited = dict([(i,False) for i in mesh.boundary_vertices])
    borders = []
    for v in mesh.boundary_vertices:
        if not visited[v]:
            border_v, _ = extract_border_cycle(mesh, v)
            for v2 in border_v:
                visited[v2] = True
            borders.append(border_v)
    return borders

@allowed_mesh_types(SurfaceMesh)
def extract_boundary_of_surface(mesh : SurfaceMesh) -> PolyLine :
    """Returns the boundary of a surface mesh as a polyline

    Args:
        mesh (SurfaceMesh): input mesh

    Returns:
        PolyLine: boundary curves of the mesh
        dict: maps a vertex id in the boundary to its corresponding id in the original mesh.
    """
    bound = PolyLine()
    visited = Attribute(bool)
    component = bound.vertices.create_attribute("component", int)
    map_v2v = dict() # vertices have a new and different index in the boundary mesh (so that indexes are [0, n])

    ind_component = 0
    ind_vertex = 0
    for v in mesh.boundary_vertices:
        if not visited[v]:
            cycle_v, cycle_e = extract_border_cycle(mesh, v)
            bound.edges += [mesh.edges[e] for e in cycle_e]
            for v2 in cycle_v:
                visited[v2] = True
                map_v2v[v2] = ind_vertex
                ind_vertex += 1
                component[v2] = ind_component
                bound.vertices.append(mesh.vertices[v2])
            ind_component += 1

    # re order edge indexes
    for e,(A,B) in enumerate(bound.edges):
        bound.edges[e] = keyify(map_v2v[A], map_v2v[B])
    return bound, map_v2v

@allowed_mesh_types(VolumeMesh)
def extract_boundary_of_volume(mesh : VolumeMesh) -> SurfaceMesh :
    """_summary_

    Args:
        mesh (VolumeMesh): _description_

    Returns:
        SurfaceMesh: the boundary as a SurfaceMesh
        dict: maps a vertex id in the boundary to its corresponding id in the original mesh.
        dict: maps a vertex id from the original mesh to its corresponding id in the boundary mesh.
    """
    bound = RawMeshData()
    map_m2b = dict() # vertices have a new and different index in the boundary mesh (so that indices are [0, n-1])
    map_b2m = dict() # boundary to mesh
    vertex_set = set()
    for iF in mesh.boundary_faces:
        bound.faces.append(iF)
        for v in mesh.faces[iF]:
            vertex_set.add(v)

    # re order vertices
    for i,v in enumerate(vertex_set):
        bound.vertices.append(mesh.vertices[v])
        map_m2b[v] = i
        map_b2m[i] = v
    # apply ordering to faces
    for i, iF in enumerate(bound.faces):
        bound.faces[i] = tuple(( map_m2b[v] for v in mesh.faces[iF]))
    bound.prepare() # ordering will be propagated to edges
    return SurfaceMesh(bound), map_m2b, map_b2m



from .base import FrameField
from ...mesh.datatypes import *
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ... import geometry as geom
from ... import operators

from ... import utils
from ... import optimize

from ...attributes import cotangent, angle_defects, mean_edge_length
from ..features import FeatureEdgeDetector
from ..connection import SurfaceConnectionFaces
from .. import trees

import numpy as np
from math import pi, atan2
import cmath
import scipy.sparse as sp
from scipy.sparse import linalg
from osqp import OSQP

class _BaseFrameField2DFaces(FrameField) : 
    """
    Base class for any frame field defined on the vertices of a surface mesh. Is not meant to be instanciated as is.
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        order:int = 4, 
        feature_edges : bool = False, 
        verbose:bool=True,
        **kwargs
    ):
        super().__init__("faces", verbose=verbose)
        self.mesh : SurfaceMesh = supporting_mesh
        self.order : int = order
        self.features : bool = feature_edges # whether feature edges are enabled or not
        self.use_cotan = kwargs.get("use_cotan", True)
        self.n_smooth = kwargs.get("n_smooth", 10)
        self.smooth_attach_weight = kwargs.get("smooth_attach_weight", None)

        self.cot : Attribute = None
        self.defect : Attribute = None
        self.conn : SurfaceConnectionFaces = kwargs.get("custom_connection", None)
        self.feat : FeatureEdgeDetector = kwargs.get("custom_features", None)
        
        self.initialized = False

    def _initialize_attributes(self):
        #processing.split_double_boundary_edges_triangles(self.mesh) # A triangle has only one edge on the boundary
        self.cot = cotangent(self.mesh, persistent=False)
        self.defect = angle_defects(self.mesh,persistent=False, dense=True)
        if self.feat is None:
            self.feat = FeatureEdgeDetector(only_border = not self.features, verbose=self.verbose)(self.mesh)
        if self.conn is None:
            self.conn = SurfaceConnectionFaces(self.mesh, self.feat)

    def _initialize_variables(self):
        self.var = np.zeros(len(self.mesh.faces), dtype=complex)

        # fix orientation on features
        for e in self.feat.feature_edges:
            e1,e2 = self.mesh.edges[e] # the edge on border
            edge = self.mesh.vertices[e2] - self.mesh.vertices[e1]
            for T in self.mesh.connectivity.edge_to_faces(e1,e2):
                if T is None: continue # edge may be on boundary
                X,Y = self.conn.base(T)
                c = complex(edge.dot(X), edge.dot(Y)) # compute edge in local basis coordinates (edge.dot(Z) = 0 -> complex number for 2D vector)
                self.var[T] = (c/abs(c))**4 # c^4 is the same for all four directions of the cross

    def _compute_attach_weight(self, A, fail_value=1e-3):
        # A is area weight matrix
        lap_no_pt = operators.laplacian_triangles(self.mesh, cotan=self.use_cotan)
        try:
            eigs = sp.linalg.eigsh(lap_no_pt, k=2, M=A, which="SM", tol=1e-3, maxiter=100, return_eigenvectors=False)
        except Exception as e:
            try:
                self.log("First estimation of alpha failed: {}".format(e))
                lap_no_pt = operators.laplacian_triangles(self.mesh, cotan=False)
                eigs = sp.linalg.eigsh(lap_no_pt+1e-3*sp.identity(lap_no_pt.shape[0]), M=A, k=2, which="SM", tol=1e-3, maxiter=100, return_eigenvectors=False)
            except:
                self.log("Second estimation of alpha failed: taking alpha = ", fail_value)
                return fail_value
        eigs_non_zero = [e for e in eigs if abs(e)>1e-6]
        if len(eigs_non_zero)==0:
            return fail_value
        return abs(min(eigs_non_zero))

    def flag_singularities(self, singul_attr_name:str = "singuls"):        
        """
        Detects singularities of the frame field

        Creates 2 attributes:
            - An attribute "singuls" on vertices storing the value (+- 1) of singularities (eventually 0 for non singular vertices)
            - An attribute "angles" on edges storing the angle of the edge, given as the difference between the two frames (associated 1-form of the frame field)
        
        Args:
            singul_attr_name (str, optional): Name of the singularity attribute created. Defaults to "singuls".
        """
        self._check_init()
        ZERO_THRESHOLD = 1e-3

        if self.mesh.edges.has_attribute("angles"):
            edge_rot = self.mesh.edges.get_attribute("angles")
            edge_rot.clear()
        else:
            edge_rot = self.mesh.edges.create_attribute("angles", float, 1, dense=True)
            # the rotation induced by the frame field on every edge
            # if edge is uv, positive orientation is from T(uv) to T(vu)

        for ie,(A,B) in enumerate(self.mesh.edges):
            T1,T2 = self.mesh.connectivity.edge_to_faces(A,B)
            if T1 is None or T2 is None: continue
            f1,f2 = self.var[T1], self.var[T2] # representation complex for T1 and T2
            
            # parallel transport
            E = self.mesh.vertices[B] - self.mesh.vertices[A]
            X1,Y1 = self.conn.base(T1)
            X2,Y2 = self.conn.base(T2)
            a1 = atan2(Y1.dot(E), X1.dot(E))
            a2 = atan2(Y2.dot(E), X2.dot(E))

            # matching
            u2 = utils.maths.roots(f2, self.order)[0]
            angles = [utils.maths.angle_diff( cmath.phase(u2) - a2, cmath.phase(u1) - a1) for u1 in utils.maths.roots(f1, self.order)]
            abs_angles = [abs(_a) for _a in angles]
            i_angle = np.argmin(abs_angles)
            edge_rot[ie] = angles[i_angle]

        if self.mesh.vertices.has_attribute(singul_attr_name):
            singuls = self.mesh.vertices.get_attribute(singul_attr_name)
            singuls.clear()
        else:
            singuls = self.mesh.vertices.create_attribute(singul_attr_name, float)
        
        for v in self.mesh.id_vertices:
            angle = self.defect[v]
            for e in self.mesh.connectivity.vertex_to_edges(v):
                u = self.mesh.connectivity.other_edge_end(e,v)
                angle += edge_rot[e] if u<v else -edge_rot[e]
            if abs(angle)>ZERO_THRESHOLD:
                singuls[v] = angle*2/pi

    def export_as_mesh(self) -> PolyLine:
        """
        Exports the frame field as a set of crosses on each faces, for visualization purposes

        Returns:
            PolyLine: representation of the frame field
        """
        FFMesh = PolyLine()
        L = mean_edge_length(self.mesh)/3
        n = self.order+1
        for id_face, face in enumerate(self.mesh.faces):
            basis,Y = self.conn.base(id_face)
            normal = geom.cross(basis,Y)
            pA,pB,pC = (self.mesh.vertices[_v] for _v in face)
            angle = cmath.phase(self.var[id_face])/self.order
            bary = (pA+pB+pC)/3 # reference point for display
            cmplx = [geom.rotate_around_axis(basis, normal, angle + 2*k*pi/self.order) for k in range(self.order)]
            pts = [bary + abs(self.var[id_face])*L*r for r in cmplx]
            FFMesh.vertices.append(bary)
            FFMesh.vertices += pts
            FFMesh.edges += [(n*id_face, n*id_face+k) for k in range(1,n)]
        return FFMesh

class FrameField2DFaces(_BaseFrameField2DFaces) : 

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        order:int = 4,  
        feature_edges : bool = False, 
        verbose:bool=True,
        **kwargs):
        super().__init__(supporting_mesh, order, feature_edges, verbose, **kwargs)
           
    def initialize(self):
        self._initialize_attributes()
        self._initialize_variables()
        self.initialized = True

    def optimize(self):
        self._check_init()
    
        self.log("Build laplacian operator")
        lap = operators.laplacian_triangles(
            self.mesh, 
            cotan=self.use_cotan, 
            connection=self.conn, 
            order=self.order).tocsc()
        A = operators.area_weight_matrix_faces(self.mesh).tocsc()

        ###### Border ######
        if len(self.feat.feature_vertices)>0: # We have a border / feature elements -> linear solve
            # Build fixed and variable indexes
            self.log("Feature element detected (border and/or feature edges)")

            fixed = self.mesh.faces.create_attribute("fixed", bool)
            for ie in self.feat.feature_edges:
                u,v = self.mesh.edges[ie]
                T1,T2 = self.mesh.connectivity.edge_to_faces(u,v)
                if T1 is not None: fixed[T1] = True
                if T2 is not None: fixed[T2] = True
            freeInds,fixedInds = [],[]
            for T in self.mesh.id_faces:
                if fixed[T] : fixedInds.append(T)
                else: freeInds.append(T)
              
            if len(freeInds)==0 : 
                self.log("Everything is on boundary : no optimization required")
                return

            lapI = lap[freeInds,:][:,freeInds]
            lapB = lap[freeInds,:][:,fixedInds]
            # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
            AI = A[freeInds, :][:, freeInds].astype(complex)
            valB = lapB.dot(self.var[fixedInds]) # right hand side
            
            self.log("Initial solve of linear system")
            res = linalg.spsolve(lapI, -valB) # first system solved without diffusion
            self.var[freeInds] = res

            if self.n_smooth>0:
                self.log(f"Solve linear system {self.n_smooth} times with diffusion")
                alpha = self.smooth_attach_weight or self._compute_attach_weight(A)
                self.log("Attach weight: {}".format(alpha))
                mat = lapI - alpha * AI
                for _ in range(self.n_smooth):
                    self.normalize()
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, -valB - valI2)
                    self.var[freeInds] = res
            self.normalize()

        ###### No border ######
        else:
            self.log("No border detected")
            self.log("Initial solve of linear system using an eigensolver")
            self.var = optimize.inverse_power_method(lap)
            if self.n_smooth>0:
                self.log(f"Solve linear system {self.n_smooth} times with diffusion")
                alpha = self.smooth_attach_weight or self._compute_attach_weight(A)
                self.log("Attach weight: {}".format(alpha))
                mat = lap - alpha * A
                solve = sp.linalg.factorized(mat)
                for _ in range(self.n_smooth):
                    self.normalize()
                    valI2 = alpha * A.dot(self.var)
                    self.var = solve(-valI2)
            self.normalize()

class TrivialConnectionFaces(_BaseFrameField2DFaces):
    """
    Implementation of 'Trivial Connections on Discrete Surfaces' by Keenan Crane and Mathieu Desbrun and Peter Schröder, 2010
    
    A frame field on faces that computes the smoothest possible frame field with prescribed singularity cones at some vertices.
    Does not constraint non-contractible cycles
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, 
        supporting_mesh : SurfaceMesh, 
        singus_indices : Attribute, 
        order : int = 4, 
        verbose : bool=True,
        **kwargs):
        super().__init__(supporting_mesh, order, feature_edges=False, verbose=verbose, **kwargs)
        self.singus = singus_indices
        self.rotations : np.ndarray = None

    def initialize(self):
        self._initialize_attributes() # /!\ may change mesh combinatorics near boundary
        self.var = np.zeros(len(self.mesh.faces), dtype=complex)
        self.initialized = True

    def optimize(self):
        self._check_init()
        ### Optimize for rotations between frames
        n_cstr = len(self.mesh.vertices)
        n_rot = len(self.mesh.edges)
        CstMat = sp.lil_matrix((n_cstr,n_rot))
        CstX = np.zeros(n_cstr)
        r = 0
        for v in self.mesh.interior_vertices:
            for e in self.mesh.connectivity.vertex_to_edges(v):
                v2 = self.mesh.connectivity.other_edge_end(e,v)
                CstMat[r,e] = 1 if v<v2 else -1
            CstX[r] = self.defect[v] - self.singus[v] * 2 * pi / self.order
            r += 1

        for v in self.mesh.boundary_vertices:
            for v2 in self.mesh.connectivity.vertex_to_vertices(v):
                if self.mesh.is_edge_on_border(v,v2): continue
                e = self.mesh.connectivity.edge_id(v,v2)
                CstMat[r,e] = 1 if v<v2 else -1
            CstX[r] = self.defect[v] - (2 - self.feat.corners[v]) * 2 * pi / self.order
            r += 1
        
        instance = OSQP()
        instance.setup(P = sp.eye(n_rot,format="csc"), q=None, A=CstMat.tocsc(), l=CstX, u=CstX, verbose=self.verbose)
        res = instance.solve()
        if res.info.status != "solved":
            raise Exception(f"Solver exited with status '{res.info.status}'. Check validity of singularity indices (Poincarré-Hopf condition)")
        self.rotations = res.x

        ### Now rebuild frame field along a tree
        for _e0 in self.feat.feature_edges : break # get a feature edge
        root = [T for T in self.mesh.connectivity.edge_to_faces(*self.mesh.edges[_e0]) if T is not None][0]
        tree = trees.FaceSpanningTree(self.mesh, root)()
        for face,parent in tree.traverse():
            if parent is None: # root
                self.var[face] = complex(1., 0.)
                continue
            zp = self.var[parent]
            ea,eb = self.mesh.connectivity.common_edge(parent,face)
            ea,eb = min(ea,eb),max(ea,eb)
            e = self.mesh.connectivity.edge_id(ea,eb)
            X1,Y1 = self.conn.base(parent)
            X2,Y2 = self.conn.base(face)
            # T1 -> T2 : angle of e in basis of T2 - angle of e in basis of T1
            E = self.mesh.vertices[eb] - self.mesh.vertices[ea]
            angle1 = atan2( geom.dot(E,Y1), geom.dot(E,X1))
            angle2 = atan2( geom.dot(E,Y2), geom.dot(E,X2))
            pt = utils.maths.principal_angle(angle2 - angle1)
            w = self.rotations[e] if self.mesh.connectivity.direct_face(ea,eb)==parent else -self.rotations[e]
            zf = zp * cmath.rect(1, 4*(w + pt))
            self.var[face] = zf
        self.smoothed = True


from abc import abstractmethod
from ..worker import Worker

class FrameField(Worker):
    """
    Abstract base class for any frame field implementations.
    A frame field is a set of vector fields defined on a surface or volume mesh.
    """

    def __init__(self, element:str, name:str="FrameField", verbose=True):
        super().__init__(name, verbose)
        self.initialized : bool = False
        self.smoothed : bool = False
        self._element : str = element
        self.var = None

    @property
    def element(self) -> str:
        """The element of a mesh on which the frame field is defined. Either vertices, edges, faces or cells"""
        return self._element

    def _check_init(self):
        if not self.initialized:
            raise Exception("FrameField was not initialized properly. Call `initialize()` before `optimize()`")

    def __getitem__(self, i):
        if self.var is not None:
            return self.var[i]
        return None

    @abstractmethod
    def initialize(self, *args, **kwargs):
        pass

    @abstractmethod
    def optimize(self, *args, **kwargs):
        pass

    def run(self):
        if not self.initialized:
            self.log("Initialize")
            self.initialize()
            self.initialized = True
        if not self.smoothed:
            self.log("Optimize")
            self.optimize()
            self.smoothed = True
        self.log("Done.")

    def normalize(self):
        if self.var is None: return
        for i in range(self.var.size):
            if abs(self.var[i])>1e-10: 
                self.var[i] /= abs(self.var[i])

    @abstractmethod
    def export_as_mesh(self, *args, **kwargs):
        pass

    @abstractmethod
    def flag_singularities(self, *args, **kwargs):
        pass


from ..mesh_data import RawMeshData
from .base import Mesh
from ... import utils

class PolyLine(Mesh): 
    """
    A data structure for representing polylines.

    Attributes:
        vertices (DataContainer): the container for all vertices
        edges (DataContainer): the container for all edges
        __str__: Representation of the object and its elements as a string.
    """

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 1, data)
        self.connectivity = PolyLine._Connectivity(self) 

    def __str__(self):
        out = "mouette.mesh.Polyline object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        out += "| {} edges\n".format(len(self.edges))
        return out

    @property
    def id_vertices(self):
        """
        Shortcut for `range(len(self.vertices))`
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for `range(len(self.edges))`
        """
        return range(len(self.edges))

    class _Connectivity:
        """
        Connectivity handler class. Allows connectivity queries on the object, like neighbours of a vertex, etc.

        Connectivity is computed lazily only when needed in the code. For instance, the first call to `mesh.connectivity.vertex_to_vertices` will generate the vertex_to_vertices dictionnary, so that the next call will not perform any computation but a lookup in the array.
        """
        
        def __init__(self, master):
            self.mesh = master

            self._edge_id : dict = None # (v1, v2) -> edge | None
            self._adjV2V : dict = None # vertex -> vertex

        def clear(self):
            """
            Resets connectivity. 
            The next query in the code will regenerate internal arrays.
            """
            self._edge_id = None
            self._adjV2V = None

        def _compute_connectivity(self):
            self._adjV2V = dict([(i,set()) for i in self.mesh.id_vertices ])

            for (A,B) in self.mesh.edges:
                assert A != B # should not be possible
                self._adjV2V[A].add(B)
                self._adjV2V[B].add(A)
                    
            for U in self.mesh.id_vertices:
                self._adjV2V[U] = list(self._adjV2V[U])

        def _compute_edge_id(self):
            self._edge_id = dict()
            for iE,E in enumerate(self.mesh.edges):
                key = utils.keyify(E)
                self._edge_id[key] = iE

        def edge_id(self, V1:int, V2:int)->int:
            """
            Identifier of an edge. If `self.edges[i]` contains edges `(A,B)`, then `edge_id(A,B)=edge_id(B,A)=i`
            
            If (A,B) is not a valid edge of the mesh, returns `None`

            Args:
                V1 (int): first vertex of the edge
                V2 (int): second vertex of the edge

            Returns:
                int: the id of edge (V1,V2), or `None` if the edge does not exist.
            """
            if self._edge_id is None:
                self._compute_edge_id()
            key = utils.keyify(V1,V2)
            return self._edge_id.get(key, None)

        def other_edge_end(self, E:int, V:int) -> int:
            """
            Vertex at the opposite end of edge `E` from vertex `V`.
            
            Returns `None` if `V` is not adjacent to edge `E`

            Args:
                E (int): edge id
                V (int): vertex id

            Returns:
                int: the vertex `W` such that `E` is the edge `(V,W)`. Returns `None` if `V` is not adjacent to edge `E`
            """
            A,B = self.mesh.edges[E]
            if V==A: return B
            if V==B: return A
            return None

        ##### Vertex to elements

        def vertex_to_vertices(self, V: int) -> list:
            """
            Neighborhood of vertex `V` in terms of vertices.

            Args:
                V (int): vertex id

            Returns:
                list: list of vertices `W` such that `(V,W)` is a valid edge in the polyline.
            """
            if self._adjV2V is None : 
                self._compute_connectivity()
            return self._adjV2V[V]

        def vertex_to_edges(self, V: int) -> list:
            """
            Neighborhood of vertex `V` in terms of edges.

            Args:
                V (int): vertex id

            Returns:
                list: list of edges E such that V belongs to E.
            """
            return [ self.edge_id(V, u) for u in self.vertex_to_vertices(V)]


        ##### Edge to elements

        def edge_to_vertices(self, E: int) -> list:
            """Returns the two vertex indices that are adjacent to edge `E

            Args:
                E (int): edge index

            Note:
                Equivalent to `mesh.edges[E]`

            Returns:
                list: two vertex indices
            """
            return self.mesh.edges[E]


from .interpolate import scatter_vertices_to_corners, interpolate_faces_to_vertices, interpolate_vertices_to_faces, average_corners_to_vertices

from .misc_vertices import degree, angle_defects, vertex_normals, border_normals
from .misc_faces import face_area, face_normals, face_circumcenter, face_barycenter, faces_near_border, triangle_aspect_ratio, parallel_transport_curvature
from .misc_corners import corner_angles, cotangent 
from .misc_edges import edge_length, curvature_matrices, cotan_weights
from .misc_cells import cell_barycenter, cell_faces_on_boundary, cell_volume

from .glob import euler_characteristic, mean_edge_length, mean_cell_volume, mean_face_area, total_area, barycenter

from .uv_export import generate_uv_colormap_corners, generate_uv_colormap_faces, generate_uv_colormap_vertices


from ..mesh_data import RawMeshData
from ..datatypes.type_checks import *
from collections import deque

def import_tet(path : str):
    """
    Read a .tet file : 
		/////////////////
        n vertices
        m tets
        v1_x v1_y v1_z
        v2_x v2_y v2_z
        ...
        vn_x vn_y vn_z
        4 v1_1 v1_2 v1_3 v1_4
        4 v2_1 v2_2 v2_3 v2_4
        ...
        4 vm_1 vm_2 vm_3 vm_3					
        /////////////////
    Parameters:
        path (str): the input file path
    """

    with open(path, 'r' ) as file:
        output = parse_tet_data(file.readlines())
    return output

def parse_tet_data(data):
    data = deque(data)
    get_line = lambda : data.popleft().strip().split()
    output = RawMeshData()
    nvert = int(get_line()[0])
    ntet  = int(get_line()[0])
    for _ in range(nvert):
        v = [float(u) for u in get_line()]
        output.vertices.append(v)
    for _ in range(ntet):
        c = tuple((int(u) for u in get_line()[1:]))
        output.cells.append(c)
    return output

def export_tet(mesh, path):
    with open( path, 'w' ) as ofile:
        ofile.write("{} vertices\n".format(len(mesh.vertices)))
        ofile.write("{} tets\n".format(len(mesh.cells)))

        for vtx in mesh.vertices:
            ofile.write(' '.join(['{}'.format(v) for v in vtx])+'\n')
        for cell in mesh.cells:
            str_face = "{} ".format(len(cell))
            str_face += " ".join([str(v) for v in cell])
            ofile.write(str_face + "\n")


from .distortion import ParamDistortion, QuadQuality
from .lscm import LSCM
from .cotan_emb import CotanEmbedding
from .tutte import TutteEmbedding
from .cone_param import ConformalConeParametrization
from .ff_integration import FrameFieldIntegration
from .winslow import WinslowInjectiveEmbedding, untangle
from .bff import BoundaryFirstFlattening


import numpy as np
from math import pi, cos, sin
from ..geometry import cotan, Vec, angle_3pts, rotate_2d
from ..mesh.datatypes import SurfaceMesh
from ..mesh.mesh import _instanciate_raw_mesh_data
from ..mesh.mesh_data import RawMeshData

def ring(N : int, defect : float, open :bool = False, n_cover:int = 1) -> SurfaceMesh:
    """
    Computes a ring of triangles with prescribed number of triangles and angle defect at the center.

    Parameters:
        N (int): number of triangles in the ring
        defect (float): target angle defect to achieve. Position of the central point is adjusted via dichotomy to match this value.
        open (bool) : whether to connect the last vertex to the first.
        n_cover (int, optional): Number of covering of the ring. Defaults to 1.

    Raises:
        Exception: Fails if N<3

    Returns:
        SurfaceMesh: the ring
    """

    ring = RawMeshData()
    max_defect = 2*pi-0.01
    defect = max(min(defect,max_defect), 0.) # 0 is ok, but 2pi is point at infinity

    if N<3:
        raise Exception("N should be > 3 for a valid ring. Aborting")

    ring.vertices.append(Vec(0.,0.,0.))
    ring.vertices.append(Vec(1.,0.,0.))

    for i in range(1, N*n_cover):
        ring.vertices.append(Vec( cos(2*i*pi/N), sin(2*i*pi/N), 0.))
        nxt = i+1 if open else (i+1)%(N*n_cover+1)
        ring.faces.append((0, i, nxt))
    if open:
        ring.vertices.append(ring.vertices[1])
        ring.faces.append((0, N*n_cover, N*n_cover+1))
    else:
        ring.faces.append((0, N*n_cover, 1))

    # Determine position of center vertex through dichotomy
    P1 = Vec(0.,0.,0.)
    P2 = Vec(0.,0.,10.)
    A = ring.vertices[1]
    B = ring.vertices[2]
    stop = False
    while not stop:
        dfct1 = 2*pi - N*angle_3pts(A,P1,B)
        dfct2 = 2*pi - N*angle_3pts(A,P2,B)

        Pmid = (P1 + P2)/2
        middfct = 2*pi - N*angle_3pts(A,Pmid,B)
        if defect>dfct2:
            P1 = P2
            P2 = 2*P2
        elif defect>middfct:
            P1 = Pmid
        elif defect<middfct:
            P2 = Pmid
        stop = (abs(dfct1 - dfct2) < 1e-6)
    ring.vertices[0] = (P1 + P2)/2
    return _instanciate_raw_mesh_data(ring, 2)

def flat_ring(N : int, defect : float, n_cover:int = 1) -> SurfaceMesh:
    ring = RawMeshData()
    max_defect = 2*pi-0.01
    defect = max(min(defect,max_defect), 0.) # 0 is ok, but 2pi is point at infinity

    ring.vertices.append(Vec(0.,0.,0.))
    
    dir = Vec(1.,0.,0.)
    ring.vertices.append(dir)
    ring.edges.append((0,1))
    ang = (2*pi-defect)/N
    for i in range(N*n_cover):
        dir = rotate_2d(dir, ang)
        dir = Vec(dir.x, dir.y, 0.)
        ring.vertices.append(dir)
        ring.faces.append((0,i+1,i+2))
    return _instanciate_raw_mesh_data(ring,2)



from .base import BaseParametrization
from ...mesh.datatypes import *
from ...mesh.mesh_attributes import *
from ...mesh.mesh import reorder_vertices

from ... import attributes
from ... import geometry as geom
from ...operators import laplacian
from ..border import extract_border_cycle
from ...attributes.glob import euler_characteristic
from ..border import extract_border_cycle

import numpy as np
import cmath
import scipy.sparse as sp

class BoundaryFirstFlattening(BaseParametrization):
    """
    Boundary First Flattening: A conformal flattening algorithm with control over the boundary 
    conditions, either in terms of scale factor or curvature

    References:
        - [1] _Boundary First Flattening_, Rohan Sawhney and Keenan Crane, ACM ToG, 2017

    Warning:
        This algorithm reorders the vertices so that boundary vertices are labeled [0, N-1] in order. Therefore, the uvs coordinates are not computed on the original mesh but on a reordered copy. Access the final result with `self.mesh`
    
    Example:
        [https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/bff.py](https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/bff.py)
    """

    def __init__(self, 
        mesh: SurfaceMesh,
        bnd_scale_fctr : Attribute = None,
        bnd_curvature : Attribute = None,
        verbose:bool=False, 
        **kwargs):
        """
        Args:
            mesh (SurfaceMesh): the input surface. Should be a triangulation of a topological disk.
            bnd_scale_fctr (Attribute, optional): Scale factor on the boundary. Ignores values for interior vertices. Defaults to None. If provided, will automatically compute boundary curvature and ignore the `bnd_curvature`a argument.
            bnd_curvature (Attribute, optional): Geodesic curvature on the boundary. Ignores values for interior vertices. Defaults to None. If provided (and no `bnd_scale_fctr` is provided), will automatically compute boundary scale factors.
            verbose (bool, optional): verbose mode. Defaults to False.
        
        Keyword Args:
            save_on_corners (bool, optional): if True, stores the results on face corners instead of vertices. Defaults to True
            use_cotan (bool, optional): if True, uses the cotan Laplacian instead of the connectivity Laplacian. Defaults to True.
            hilbert_transform (bool, optional): if True, extends the boundary values with the Hilbert transform 
            (favors conformality instead of boundary preservation) instead of using harmonic extensions 
            (exact boundary but further from truly conformal). Defaults to True.

        Attributes:
            uvs (Attribute): an attribute containing the uv-coordinates of the parametrization

        Raises:
            Exception: if the mesh is not the triangulation of a topological disk

        Note:
            If neither of `bnd_scale_fctr` and `bnd_curvature` are provided, the algorithm will run in default mode with scale factors = 0 on the boundary. Otherwise, provided scale factors have priority over provided curvatures.
        """

        super().__init__("BFF", mesh, verbose, **kwargs)
        self._Bscale_fctr = bnd_scale_fctr
        self._Bcurvature = bnd_curvature
        self.use_cotan = kwargs.get("use_cotan", True)
        self.hilbert_transform = kwargs.get("hilbert_transform", True)

        self._curv : ArrayAttribute = None
        self.NV,self.NB = None,None
        self._L : sp.csc_matrix = None # Laplacian operator
        self._Lii, self._Lib, self._Lbb = None, None, None # Laplacian submatrices
        
        self._vertex_permutation = None
        self._Bedges = None # list of edges (in order) that make the boundary
        self._Blengths = None
        self._Btarget_lenghts = None

    def run(self) -> SurfaceMesh:
        """Run the algorithm

        Raises:
            Exception: If the mesh is not a triangulation of a topological disk

        Returns:
            SurfaceMesh: a copy of the original mesh with reordered vertices and computed uv-coordinates
        """
        if not (len(self.mesh.boundary_vertices)>1 and euler_characteristic(self.mesh)==1):
            raise Exception("Mesh is not a topological disk. Cannot run parametrization.")
        if not self.mesh.is_triangular():
            raise Exception("Mesh is not a triangulation. Cannot run parametrization.")

        self.log("Reorder mesh vertices")
        self._Bedges = self._reorder_vertices()
        self.NV, self.NB = len(self.mesh.vertices), len(self.mesh.boundary_vertices)

        self.log("Compute Gaussian and geodesic curvatures")
        self._curv  = attributes.angle_defects(self.mesh).as_array(self.NV) # Gaussian curvature

        self.log("Compute Laplacian operator")
        self._L = laplacian(self.mesh, cotan = self.use_cotan)

        self._Lii = self._L[ self.NB:,:][:, self.NB:]
        self._Lib = self._L[ self.NB:,:][:,:self.NB ]
        self._Lbb = self._L[:self.NB ,:][:,:self.NB ]

        self.log("Initialize Boundary scale factor and curvature")
        if self._Bscale_fctr is None and self._Bcurvature is None:
            # Nothing was provided. By default, we run with zero scale factor on the boundary
            self._Bscale_fctr = np.zeros(self.NB)
        
        if self._Bscale_fctr is not None:
            self._Bscale_fctr = np.array([self._Bscale_fctr[v] for v in self._vertex_permutation[:self.NB]])
            # Curvature was provided -> extrapolate scale factors via Neumann to Dirichlet
            self._Bcurvature = self._dirichlet_to_neumann(-self._curv, self._Bscale_fctr)
        elif self._Bcurvature is not None:
            # Scale factors were provided -> extrapolate curvature via Dirichlet to Neumann
            self._Bscale_fctr = self._neumann_to_dirichlet()
        else:
            raise Exception("Should not happen.")
                
        self.log("Fit best boundary curve")
        Ub,Vb = self._fit_boundary_curve()

        self.log("Extend curve values inside the surface")
        U,V = self._extend_boundary_values(Ub,Vb)

        self.log("Write UV-coordinates")
        if self.save_on_corners:
            self.uvs = self.mesh.face_corners.create_attribute("uv_coords", float, 2, dense=True)
            for ic,v in enumerate(self.mesh.face_corners):
                self.uvs[ic] = Vec(U[v], V[v])
        else:
            self.uvs = self.mesh.vertices.register_array_as_attribute("uv_coords", np.vstack((U,V)).T)
        return self.mesh

    def _reorder_vertices(self):
        v_border, _ = extract_border_cycle(self.mesh)
        self._vertex_permutation = v_border + self.mesh.interior_vertices
        self.mesh = reorder_vertices(self.mesh, self._vertex_permutation)
        _, e_border = extract_border_cycle(self.mesh)
        return e_border

    def _dirichlet_to_neumann(self, phi, U):
        a = sp.linalg.spsolve(self._Lii, phi[self.NB:] - self._Lib @ U)
        return phi[:self.NB] - self._Lib.transpose() @ a - self._Lbb @ U

    def _neumann_to_dirichlet(self):    
        raise NotImplementedError

    def _fit_boundary_curve(self):
        self._Blengths = np.zeros(len(self.mesh.boundary_edges))
        self._Btarget_lenghts = np.zeros(len(self.mesh.boundary_edges))
        for ie,e in enumerate(self._Bedges):
            A,B = self.mesh.edges[e]
            uA,uB = self._Bscale_fctr[A], self._Bscale_fctr[B]
            self._Blengths[ie] = geom.distance( self.mesh.vertices[A], self.mesh.vertices[B])
            self._Btarget_lenghts[ie] = self._Blengths[ie] * np.exp((uA+uB)/2)
        
        T = np.zeros((self.NB,2))
        direction = 0.
        for i in range(self.NB):
            direction += self._Bcurvature[i]
            T[i] = geom.Vec.from_complex(cmath.rect(1, direction))
        N = sp.diags(self._Blengths)
        final_lengths = self._Btarget_lenghts - N @ T @ np.linalg.inv(T.transpose() @ N @ T) @ T.transpose() @ self._Btarget_lenghts
        bnd_pts = np.zeros((self.NB,2))
        for i in range(1,self.NB):
            bnd_pts[i] = bnd_pts[i-1] + final_lengths[i-1] * T[i-1]
        return bnd_pts[:,0], bnd_pts[:,1]
    
    def _extend_boundary_values(self, Ub, Vb):
        if self.hilbert_transform:
            Ui = sp.linalg.spsolve(self._Lii, - self._Lib @ Ub) # extending u coordinate
            U = np.concatenate((Ub,Ui))
            h = np.zeros(self.NV)
            for i in range(self.NB):
                h[i] = (Ub[(i+1)%self.NB] - Ub[(i-1)%self.NB])/2
            V = sp.linalg.spsolve(self._L, h)
        else:
            Ui = sp.linalg.spsolve(self._Lii, - self._Lib @ Ub) # extending u coordinate
            Vi = sp.linalg.spsolve(self._Lii, -self._Lib @ Vb) # extending v coordinate
            U,V = np.concatenate((Ub,Ui)), np.concatenate((Vb,Vi))    
        return U,V


import numpy as np

from ..geometry import *
from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from ..mesh.mesh import _instanciate_raw_mesh_data
from ..mesh.subdivision import SurfaceSubdivision
from .dual import dual_mesh
from scipy.spatial import ConvexHull

from math import sqrt, pi, cos, sin

def tetrahedron(P1: Vec, P2: Vec, P3: Vec, P4: Vec, volume: bool=False) -> SurfaceMesh:
    """Simple tetrahedron from four points

    Args:
        P1 (Vec): first point
        P2 (Vec): second point
        P3 (Vec): third point
        P4 (Vec): fourth point
        volume (bool, optional): whether to generate a cell or just a surface. Defaults to False.

    Returns:
        SurfaceMesh: _description_
    """
    tet = RawMeshData()
    tet.vertices += [P1,P2,P3,P4]
    tet.faces += [(1,2,3), (0,2,3),(0,1,3),(0,1,2)]
    if volume: tet.cells.append((0,1,2,3))
    return _instanciate_raw_mesh_data(tet)

def hexahedron(
        P1: Vec, P2: Vec, P3: Vec, P4: Vec, P5: Vec, P6: Vec, P7: Vec, P8: Vec, 
        colored: bool = False, triangulate: bool = False, volume: bool = False
    ) -> SurfaceMesh:
    """Generate an hexahedron in arbitrary configuration given 8 points. Order and connectivity of points is:

    ```
       7--------6
      /|       /|
     / |      / |
    4--------5  |
    |  |     |  |
    |  3-----|--2
    | /      | /
    |/       |/
    0--------1
    ```

    Parameters:
        P1 to P8 (Vec): coordinates of eight vertices
        colored (bool, optional): if set to true, will add a color attribute on faces. Defaults to False.
        triangulate (bool, optional): if set to true, will triangulate the faces. Defaults to False.
        volume (bool, optional): if set to true, will also generate three tetrahedra to fill the volume. Defaults to False.

    Returns:
        SurfaceMesh: a cube
    """
    hexa = RawMeshData()
    hexa.vertices += [P1,P2,P3,P4,P5,P6,P7,P8]

    if volume:
        #hexa.cells += [(0,1,2,3),(0,5,7,4), (2,6,7,5)]
        hexa.cells += [(0,1,2,3,4,5,6,7)]
    else:
        if triangulate:
            hexa.faces += [
                (0,2,1), (0,3,2),
                (0,1,5), (0,5,4),
                (1,2,6), (1,6,5),
                (2,3,7), (2,7,6),
                (3,0,4), (3,4,7),
                (4,5,6), (4,6,7)
            ]
        else:
            hexa.faces += [
                (0,3,2,1),
                (0,1,5,4),
                (1,2,6,5),
                (2,3,7,6),
                (3,0,4,7),
                (4,5,6,7)
            ]
        if colored:
            col = hexa.faces.create_attribute("color", float, 3)
            RED,GREEN,BLUE = Vec(1.,0.,0), Vec(0.,1.,0.), Vec(0.,0.,1.)
            col[0] = RED
            col[1] = RED
            col[10] = RED
            col[11] = RED

            col[2] = GREEN
            col[3] = GREEN
            col[6] = GREEN
            col[7] = GREEN
            
            col[4] = BLUE
            col[5] = BLUE
            col[8] = BLUE
            col[9] = BLUE
    return _instanciate_raw_mesh_data(hexa)

def axis_aligned_cube(colored: bool = False, triangulate: bool = False) -> SurfaceMesh:
    """generated an axis aligned cube as 6 quad faces.

    ```
       7--------6
      /|       /|
     / |      / |
    4--------5  |
    |  |     |  |
    |  3-----|--2
    | /      | /
    |/       |/
    0--------1
    ```

    Parameters:
        colored (bool, optional): if set to true, will add a colo rattribute on faces to determine. Defaults to False.
        triangulate (bool, optional): if set to true, will triangulate the faces. Defaults to False.

    Returns:
        SurfaceMesh: a cube
    """
    v0 = Vec(-0.5,-0.5,-0.5)
    v1 = Vec(0.5,-0.5,-0.5)
    v2 = Vec(0.5,0.5,-0.5)
    v3 = Vec(-0.5,0.5,-0.5)

    v4 = Vec(-0.5,-0.5,0.5)
    v5 = Vec(0.5,-0.5,0.5)
    v6 = Vec(0.5,0.5,0.5)
    v7 = Vec(-0.5,0.5,0.5)
    return hexahedron(v0,v1,v2,v3,v4,v5,v6,v7, colored=colored, triangulate=triangulate)

def hexahedron_4pts(P1: Vec, P2: Vec, P3: Vec, P4: Vec, colored=False, volume=False) -> SurfaceMesh:
    """Generate an hexahedron given by an absolute position and three points building a basis.

    ```
    4
    |
    |  3
    | /
    |/
    1--------2
    ```

    Parameters:
        P1 to P4 (Vec): coordinates of vertices
        colored (bool, optional): if set to true, will add a color attribute on faces to determine. Defaults to False.
        volume (bool, optional): if set to true, will also generate three tetrahedra to fill the volume. Defaults to False.

    Returns:
        SurfaceMesh: [description]
    """
    X,Y = P2-P1, P3-P1
    return hexahedron(P1, P1+X, P1+X+Y, P1+Y, P4, P4+X, P4+X+Y, P4+Y, colored, volume)

def octahedron():
    """Generate a unit octahedron as the dual mesh of the unit hexahedron
    
    Reference:
        https://danielsieger.com/blog/2021/01/03/generating-platonic-solids.html
    """
    return dual_mesh(axis_aligned_cube())

def icosahedron(center : Vec = Vec(0,0,0), radius: float=1., uv=False):
    """Generate a unit icosahedron

    Args:
        center (Vec, optional): center position. Defaults to Vec(0,0,0).
        uv (bool, optional): whether to generate uv coordinates. Defaults to False.

    Returns:
        _type_: _description_
    """
    phi = (1 + sqrt(5)) / 2
    m = RawMeshData()

    m.vertices += [ radius*a+center for a in 
    [
        Vec(-1, phi,0),
        Vec(1, phi, 0),
        Vec(-1, -phi, 0),
        Vec(1, -phi, 0),

        Vec(0, -1, phi),
        Vec(0, 1, phi),
        Vec(0, -1, -phi),
        Vec(0, 1, -phi),

        Vec(phi, 0, -1),
        Vec(phi, 0, 1),
        Vec(-phi, 0, -1),
        Vec(-phi, 0, 1),
    ]]

    m.faces += [
        (0, 11, 5),
        (0, 5, 1),
        (0, 1, 7),
        (0, 7, 10),
        (0, 10, 11),
        (1, 5, 9),
        (5, 11, 4),
        (11, 10, 2),
        (10, 7, 6),
        (7, 1, 8),
        (3, 9, 4),
        (3, 4, 2),
        (3, 2, 6),
        (3, 6, 8),
        (3, 8, 9),
        (4, 9, 5),
        (2, 4, 11),
        (6, 2, 10),
        (8, 6, 7),
        (9, 8, 1)
    ]
    return SurfaceMesh(m)

def dodecahedron() -> SurfaceMesh:
    """Generate a unit dodecahedron as the dual mesh of an icosahedron

    Returns:
        SurfaceMesh: a dodecahedron
    """
    return dual_mesh(icosahedron())

def cylinder(P1 : Vec, P2 : Vec, radius: float = 1., N=50, fill_caps=True) -> SurfaceMesh:
    """Generates a cylinder around the segment defined by P1 and P2

    Args:
        P1 (Vec): start of the cylinder (center of bottom face)
        P2 (Vec): end of the cylinder (center of top face)
        radius (float, optional): radius. Defaults to 1..
        N (int, optional): number of segments. Defaults to 50.
        fill_caps (bool, optional): whether to also generate faces at caps. Defaults to True.

    Returns:
        SurfaceMesh: a cylinder
    """
    cy = RawMeshData()
    axis = Vec.normalized(P2-P1)
    t = Vec(axis.y, -axis.x, 0.) # tangent vector
    if t.norm()<1e-6:
        t = Vec(0, axis.z, -axis.y) # another tangent vector
    t = Vec.normalized(t)
    for P in (P1,P2):
        for i in range(N):
            Pi = P + radius * rotate_around_axis(t, axis, 2*np.pi*i/N)
            cy.vertices.append(Pi)
    # caps
    if fill_caps:
        cy.vertices += [P1,P2]
        for i in range(N):
            cy.faces.append((i,(i+1)%N,2*N))
            cy.faces.append((i+N,2*N+1, (i+1)%N+N))
    # side of cylinder
    for i in range(N):
        cy.faces.append((i,  N+i, (i+1)%N))
        cy.faces.append((N+i, N+(i+1)%N, (i+1)%N))
    return SurfaceMesh(cy)

def torus(
    major_segments: int = 50,
    minor_segments: int = 30,
    major_radius: float = 1.,
    minor_radius: float = 0.3,
    triangulate: bool = False
) -> SurfaceMesh:
    """Generates a torus
    Args:
        major_segments (int): number of major segments. Defaults to 50.
        minor_segments (int): number of minor segments. Defaults to 30.
        major_radius (float): global radius of the torus. Defaults to 1.
        minor_radius (float): thickness of the torus. Defaults to 0.3
        triangulate (bool, optional): whether to output a triangular or quadmesh. Defaults to False.

    References:
        https://danielsieger.com/blog/2021/05/03/generating-primitive-shapes.html

    Returns:
        SurfaceMesh: a torus
    """
    out = RawMeshData()
    # generate vertices
    for i in range(major_segments):
       for j in range(minor_segments):
            u = i / major_segments * 2 * np.pi 
            v = j / minor_segments * 2 * np.pi
            x = (major_radius + minor_radius * np.cos(v)) * np.cos(u)
            y = (major_radius + minor_radius * np.cos(v)) * np.sin(u)
            z = minor_radius * np.sin(v)
            out.vertices.append(Vec(x,y,z))

    for i in range(major_segments):
        i_next = (i+1)%major_segments
        for j in range(minor_segments):
            j_next = (j + 1) % minor_segments
            v0 = i * minor_segments + j
            v1 = i * minor_segments + j_next
            v2 = i_next * minor_segments + j_next
            v3 = i_next * minor_segments + j
            if triangulate:
                out.faces += [(v0,v1,v3), (v1,v2,v3)]
            else:
                out.faces.append((v0,v1,v2,v3))
    return _instanciate_raw_mesh_data(out, 2)

def sphere_uv( n_lat : int = 30, n_long : int= 50, center : Vec = Vec(0.,0.,0.), radius : float = 1.) -> SurfaceMesh:
    """Generates a sphere using classical spherical uv-coordinates 

    Parameters:
        n_lat (int): number of different latitudes for points. Defaults to 30.
        n_long (int): number of different longitudes for points. Defaults to 50.
        center (Vec, optional): Center position of the sphere. Defaults to Vec(0.,0.,0.).
        radius (float, optional): Radius of the sphere. Defaults to 1.

    References:
        https://danielsieger.com/blog/2021/03/27/generating-spheres.html

    Returns:
        SurfaceMesh: the sphere
    """
    sp = RawMeshData()

    # add north pole
    sp.vertices.append(center + radius*Vec(0.,0.,1.))

    # add other points
    for i in range(n_lat):
        phi = np.pi * (i+1)/n_lat
        for j in range(n_long):
            theta = 2*np.pi * j/n_long
            x = np.sin(phi)*np.cos(theta)
            y = np.sin(phi)*np.sin(theta)
            z = np.cos(phi)
            sp.vertices.append(center + radius*Vec(x,y,z))

    # add south pole
    sp.vertices.append(center + radius*Vec(0.,0.,-1.))

    # add rings of triangles at poles
    for i in range(n_long):
        i0, i1 = i+1, (i+1)%n_long+1
        sp.faces.append((i0, 0, i1))

        i0 = i + n_long * (n_lat - 2) + 1
        i1 = (i + 1) % n_long + n_long * (n_lat - 2) + 1
        sp.faces.append((len(sp.vertices)-1, i0, i1))

    # add rows of quads
    for j in range(n_lat-2):
        j0 = j*n_long + 1
        j1 = (j+1)* n_long + 1
        for i in range(n_long):
            i0 = j0 + i
            i1 = j0 + (i + 1) % n_long
            i2 = j1 + (i + 1) % n_long
            i3 = j1 + i
            sp.faces.append((i0,i1,i2,i3))
    return _instanciate_raw_mesh_data(sp,2)


def icosphere(n_refine : int= 3, center: Vec = Vec(0.,0.,0.), radius: float = 1.) -> SurfaceMesh:
    """Generates an icosphere, that is a subdivision of the icosahedron

    Args:
        n_refine (int, optional): number of subdivisions. Defaults to 3.
        center (Vec, optional): center of the sphere. Defaults to Vec(0.,0.,0.).
        radius (float, optional): radius of the sphere. Defaults to 1..

    Returns:
        SurfaceMesh: the sphere
    """
    ico = icosahedron(center, radius)
    # Subdivide
    with SurfaceSubdivision(ico,False) as subdiv:
        for _ in range(n_refine):
            subdiv.loop_subdivision(1)
            for iv in subdiv.mesh.id_vertices:
                subdiv.mesh.vertices[iv] = center + radius*Vec.normalized(subdiv.mesh.vertices[iv]-center)
    return subdiv.mesh

def sphere_fibonacci( n_pts : int, radius:float =1., build_surface : bool = True) -> SurfaceMesh:
    """Generates a point cloud or a surface mesh using fibonacci sampling of a sphere.

    Parameters:
        n_pts (int): total number of vertices
        radius (float, optional): Radius of the sphere. Defaults to 1.
        build_surface (bool, optional): If specified to True, the function will also compute a triangulation of the vertices. This is obtained through a convex hull algorithm (since points lay on a convex shape, the convex hull and the Delaunay triangulation are equivalent). Defaults to True.

    Returns:
        [SurfaceMesh | PointCloud]: the generated mesh
    """

    phi = 0.5 * (1. + np.sqrt(5.))

    theta = np.zeros(n_pts)
    sphi = np.zeros(n_pts)
    cphi = np.zeros(n_pts)

    points = []
    for i in range(n_pts):
        j = 2*i - (n_pts-1) 
        theta = 2.0 * np.pi *  j / phi
        sphi = j / n_pts
        cphi = np.sqrt( (n_pts + j ) * ( n_pts - j )) /  n_pts

        x = cphi * np.sin(theta)
        y = cphi * np.cos(theta)
        z = sphi
        points.append(radius*Vec(x,y,z))

    data = RawMeshData()
    data.vertices += points
    if build_surface:
        # Triangulate points on a sphere : Delaunay is equivalent to convex hull since every points lay in the hull
        ch = ConvexHull(points, qhull_options="QJ")
        for (A,B,C) in ch.simplices:
            # Correctly orient the face with outward normal
            pA,pB,pC = (points[_x] for _x in (A,B,C))
            pcenter = (pA+pB+pC)/3
            normal = geom.cross(pB-pA,pC-pA)
            if geom.dot(pcenter,normal)<0:
                data.faces.append((A,C,B))
            else:
                data.faces.append((A,B,C))
    return _instanciate_raw_mesh_data(data)



from random import randint
from collections import deque
from math import isinf

from ...mesh.datatypes import *
from ...mesh.mesh_attributes import Attribute
from ...utils import keyify, UnionFind
from ...attributes.misc_edges import edge_length as attr_edge_length
from .base import SpanningForest, SpanningTree

class EdgeSpanningTree(SpanningTree):
    """A spanning tree defined over the connectivity of a mesh.
    Edges of a polyline, surface mesh or volume mesh form an undirected graph, from which we can extract a spanning tree
    
    Warning:
        This tree considers a unique starting point and stops when all reachable vertices are visited. 
        If the mesh is disconnected, the tree will be incomplete. In this case, use `EdgeSpanningForest` class instead.

    Attributes:
        parent (list): the list of indices of the parent node in the tree. `None` for the root.
        children (list): list of indices of children nodes.
        edges (list): list of edges that form the tree.
    """

    @forbidden_mesh_types(PointCloud)
    def __init__(self, mesh : Mesh, starting_vertex : int = None, avoid_boundary : bool = False, avoid_edges : set = None):
        """
        Parameters:
            mesh (Mesh): the input mesh (Polyline, Surface or Volume)
            starting_vertex (int, optional): Index of the root of the tree. If None is provided, root is chosen at random. Defaults to None.
            avoid_boundary (bool, optional): If True, boundary edges of a SurfaceMesh will not be traversed by the tree. Defaults to False.
            avoid_edges (set, optional): Set of edges that should not be travelled by the tree. 
                /!\\ If this set disconnects the mesh, the tree will be incomplete. Defaults to None.
        """
        super().__init__(mesh)
        self._avoidbound : bool = avoid_boundary
        self._avoidedges : set = avoid_edges
        if starting_vertex is not None:
            self.root = starting_vertex
        else:
            self.root = randint(0,len(self.mesh.vertices)-1)
        
        self.parent = [None]*len(self.mesh.vertices)
        self.children = [[] for _ in self.mesh.id_vertices]
        self.edges = []
    

    def _avoid_edge(self, a: int, b:int) -> bool:
        """If avoid_boundary was set to True, checks if the edge is on the boundary and should not be considered in the search.
        Always returns False otherwise.
        Parameters:
            a (int): index of first vertex of edge
            b (int): index of second vertex of edge

        Returns:
            bool : whether to avoid the edge or not
        """
        if self._avoidedges is not None:
            if self.mesh.connectivity.edge_id(a,b) in self._avoidedges:
                return True
        if not self._avoidbound or isinstance(self.mesh, PolyLine):
            return False
        return self.mesh.is_edge_on_border(a,b)

    def compute(self):
        dist_to_root = [float("inf") for v in self.mesh.id_vertices]
        seen = [False for v in self.mesh.id_vertices]
        queue = deque()

        def put_neighbours_in_queue(v):
            for nv in self.mesh.connectivity.vertex_to_vertices(v):
                if (not seen[nv]) and (not self._avoid_edge(v,nv)):
                    queue.append((v,nv))
        
        put_neighbours_in_queue(self.root)
        dist_to_root[self.root] = 0
        seen[self.root] = True

        while len(queue)>0:
            v,nv = queue.popleft()
            if seen[nv] : continue
            seen[nv] = True
            if dist_to_root[v] + 1 < dist_to_root[nv]:
                self.parent[nv] = v
                dist_to_root[nv] = dist_to_root[v] +1
            put_neighbours_in_queue(nv)

        # build children data from parent data
        for v in self.mesh.id_vertices:
            if isinf(dist_to_root[v]): continue # vertex was not reached by the tree
            p = self.parent[v]
            if p is not None:
                self.children[p].append(v)
                self.edges.append( keyify(p,v))
        self._computed = True # sets the 'computed' flag

    def build_tree_as_polyline(self) -> PolyLine:
        """Builds the tree as a new polyline object. Useful for debug and visualization purposes

        Returns:
            PolyLine: the tree
        """
        output = PolyLine()
        for v in self.mesh.vertices:
            output.vertices.append(v)
        for v,father in self.traverse():
            if father is None: continue
            output.edges.append(keyify(v, father))
        return output

class EdgeMinimalSpanningTree(EdgeSpanningTree):
    """Same as EdgeSpanningTree but uses Kruskal's algorithm instead of a Breadth First Search.
    This implies that the resulting spanning tree is minimal in term of edge lengths.

    Inherits from `EdgeSpanningTree`.

    Warning:
        This tree considers a unique starting point and stops when all reachable vertices are visited. 
        If the mesh is disconnected, the tree will be incomplete. In this case, use `EdgeSpanningForest` class instead.
    """

    @forbidden_mesh_types(PointCloud)
    def __init__(self, mesh : Mesh, starting_vertex=None, avoid_boundary : bool = False, weights="length"):
        """
        Parameters:
            mesh (Mesh): the input mesh (Polyline, Surface or Volume)
            starting_vertex (int, optional): Index of the root of the tree. If None is provided, root is chosen at random. Defaults to None.
            avoid_boundary (bool, optional): If True, boundary edges will not be traversed by the tree. Defaults to False.
            weights (str|dict, optional) : provided weights of each edge. Options are:
                `one` : uniform weight = 1 for every edge;
                `length` : use the length of the edge;
                `any dict` : custom weights.
                Defaults to "length".
        """
        super().__init__(mesh, starting_vertex, avoid_boundary=avoid_boundary)
        if not ((isinstance(weights, str) and weights in ["one", "length"]) or isinstance(weights, dict) or isinstance(weights, Attribute)):
            raise Exception("Acceptable weights are 'one', 'length' or a custom dict or Attribute on edges")
        self.weights = weights

    def compute(self):
        if self.weights=="one":
            edge_length = lambda _ : 1.
        elif self.weights== "length":
            _edge_length = attr_edge_length(self.mesh, persistent=False)
            edge_length = lambda e : _edge_length[e]
        else:
            edge_length = lambda e : self.weights[e]

        if not self._avoidbound or isinstance(self.mesh, PolyLine):
            edges = [e for e in self.mesh.id_edges]
        else:
            edges = [e for e,(A,B) in enumerate(self.mesh.edges) if not self.mesh.is_edge_on_border(A,B)]

        edges.sort(key = lambda e : edge_length(e))
        neighbours = [set() for _ in self.mesh.id_vertices]
        uf = UnionFind(self.mesh.id_vertices)
        for e in edges:
            a,b = self.mesh.edges[e]
            if not uf.connected(a,b):
                uf.union(a,b)
                self.edges.append(keyify(a,b))
                neighbours[a].add(b)
                neighbours[b].add(a)

        # build parents and children from neighbours -> BFS from root
        queue = deque()
        self.parent[self.root] = None
        self.children[self.root] = list(neighbours[self.root])
        for v in neighbours[self.root]:
            queue.append((v, self.root))
        while len(queue)>0:
            v,prev= queue.popleft()
            self.parent[v] = prev
            self.children[v] = [x for x in neighbours[v] if x != prev]
            for child in self.children[v]:
                queue.append((child,v))
        self._computed = True # sets the 'computed' flag
        
class EdgeSpanningForest(SpanningForest):
    """A spanning forest that runs on edges. 
    Unlike a spanning tree, will create new roots and expand trees until all vertices of the mesh have been visited
    """

    @forbidden_mesh_types(PointCloud)
    def __init__(self, mesh : Mesh):
        super().__init__(mesh)

    def compute(self) -> None :
        visited = [False]*len(self.mesh.vertices)
        for v in self.mesh.id_vertices:
            if not visited[v]:
                # create a tree starting from v
                self.roots.append(v)
                tree_v = EdgeSpanningTree(self.mesh, v)()
                self.trees.append(tree_v)
                for (node, _) in tree_v.traverse():
                    visited[node] = True
        super().compute() # sets the 'computed' flag


from ..mesh_data import RawMeshData
from ..mesh_attributes import Attribute
from .base import Mesh
from .surface import PolyLine, SurfaceMesh
from ...utils import keyify
from ...geometry.geometry import det_3x3
from ... import config

import numpy as np

class VolumeMesh(Mesh):

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 3, data)
        self.connectivity = VolumeMesh._Connectivity(self)

        # Boundary data
        self._boundary_faces : list = None
        self._interior_faces : list = None

        self._is_vertex_on_border : Attribute = None 
        self._boundary_vertices : list = None
        self._interior_vertices : list = None

        self._is_edge_on_border : Attribute = None
        self._boundary_edges : list = None
        self._interior_edges : list = None

        self.boundary_connectivity : VolumeMesh._BoundaryConnectivity = None

    def enable_boundary_connectivity(self):
        """
        Builds the indirection maps to the boundary surface mesh, which then can be accessed via the `self.boundary_connectivity` attribute.

        See documentation for VolumeMesh._BoundaryConnectivity for more details.
        """
        self.boundary_connectivity = VolumeMesh._BoundaryConnectivity(self)

    def __str__(self):
        out = "mouette.mesh.VolumeMesh object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        out += "| {} edges\n".format(len(self.edges))
        out += "| {} faces\n".format(len(self.faces))
        out += "| {} cells\n".format(len(self.cells))
        return out

    @property
    def id_vertices(self):
        """
        Shortcut for `range(len(self.vertices))`
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for `range(len(self.edges))`
        """
        return range(len(self.edges))

    @property
    def id_faces(self):
        """
        Shortcut for `range(len(self.faces))`
        """
        return range(len(self.faces))

    @property
    def id_cells(self):
        """
        Shortcut for `range(len(self.cells))`
        """
        return range(len(self.cells))
    
    @property
    def id_corners(self):
        """
        Shortcut for `range(len(self.face_corners))`
        """
        return range(len(self.face_corners))

    @property
    def boundary_mesh(self):
        """The boundary of the VolumeMesh as a SurfaceMesh. Shortcut for `self.boundary_connectivity.mesh
        
        Returns:
            None: if the boundary connectivity is not enabled (mesh was not built). See `enable_connectivity`
            SurfaceMesh: the boundary mesh otherwise

        """
        try:
            return self.boundary_connectivity.mesh
        except Exception as _:
            return None

    def _compute_interior_boundary_faces(self):
        self._interior_faces = []
        self._boundary_faces = []
        for F in self.id_faces:
            if self.is_face_on_border(F):
                self._boundary_faces.append(F)
            else:
                self._interior_faces.append(F)

    def _compute_interior_boundary_vertices(self):
        self._is_vertex_on_border = self.vertices.create_attribute("border",bool)
        for iF in self.boundary_faces:
            for v in self.faces[iF]:
                self._is_vertex_on_border[v] = True
        self._interior_vertices = []
        self._boundary_vertices = []
        for v in self.id_vertices:
            if self._is_vertex_on_border[v]:
                self._boundary_vertices.append(v)
            else:
                self._interior_vertices.append(v)

    def _compute_interior_boundary_edges(self):
        self._is_edge_on_border = self.edges.create_attribute("border", bool)
        for iF in self.boundary_faces:
            n = len(self.faces[iF])
            for i in range(n):
                u,v = self.faces[iF][i], self.faces[iF][(i+1)%n]
                self._is_edge_on_border[self.connectivity.edge_id(u,v)] = True
        self._interior_edges = []
        self._boundary_edges = []
        for e in self.id_edges:
            if self._is_edge_on_border[e]:
                self._boundary_edges.append(e)
            else:
                self._interior_edges.append(e)

    def is_face_on_border(self, *args) -> bool:
        """Simple test to determine if a given face is on the boundary of the mesh.
        
        Parameters:
            *args: Either an integer index representing a face, or n integer indices representing the vertices
            
        Returns:
            bool: Returns True is the given face exists and is on the boundary of the mesh
        """
        if len(args)==1:
            n = len(self.connectivity.face_to_cells(args[0]))
        else:
            n = len(self.connectivity.face_to_cells(self.connectivity.face_id(*args)))
        return n<2
    
    def is_vertex_on_border(self, V) -> bool:
        if self._is_vertex_on_border is None:
            self._compute_interior_boundary_vertices()
        return self._is_vertex_on_border[V]
    
    def is_edge_on_border(self, *args) -> bool:
        """Simple test to determine if a given edge is on the boundary of the mesh.

        Parameters:
            *args: Either an integer index representing an edge, or two integer indices representing two (adjacent) vertices

        Returns:
            bool: Returns True if the given edge is on the boundary of the mesh.
        """
        if self._is_edge_on_border is None:
            self._compute_interior_boundary_edges()
        if len(args)==1:
            return self._is_edge_on_border[args[0]]
        return self._is_edge_on_border[self.connectivity.edge_id(args[0],args[1])]

    def is_cell_tet(self, ic:int) -> bool:
        """Returns `True if the cell given by id `ic` is a tetrahedron (i.e. has 4 vertices)
        
        Parameters:
            ic (int): cell index
            
        Returns:
            bool: len(self.cells[ic])==4 """
        return len(self.cells[ic])==4
    
    # def is_cell_hex(self, ic) -> bool:
    #     """Returns True if the cell given by id `ic` is a hexahedron (has 8 vertices) 
        
    #     Warning:
    #         Having 8 vertices is not a sufficient condition for a cell to be a hexahedron, though it is a good enough test.
    #     """
    #     return len(self.cells[ic])==8

    def is_tetrahedral(self) -> bool:
        """
        Returns:
            bool: True if the mesh is tetrahedral (all cells are tetrahedra)
        """
        return np.all([self.is_cell_tet(ic) for ic in self.id_cells])

    @property
    def boundary_faces(self):
        if self._boundary_faces is None:
            self._compute_interior_boundary_faces() 
        return self._boundary_faces

    @property
    def interior_faces(self):
        if self._interior_faces is None :
            self._compute_interior_boundary_faces() 
        return self._interior_faces

    @property
    def interior_edges(self):
        if self._interior_edges is None :
            self._compute_interior_boundary_edges()
        return self._interior_edges

    @property
    def boundary_edges(self):
        if self._boundary_edges is None:
            self._compute_interior_boundary_edges()
        return self._boundary_edges

    @property
    def boundary_vertices(self):
        if self._boundary_vertices is None :
            self._compute_interior_boundary_vertices()
        return self._boundary_vertices

    @property
    def interior_vertices(self):
        if self._interior_vertices is None:
            self._compute_interior_boundary_vertices()
        return self._interior_vertices

    class _Connectivity(SurfaceMesh._Connectivity): 

        def __init__(self, master):
            super().__init__(master)
            self._adjC2C : Attribute = None
            self._adjV2C : dict = None

            self._adjC2F : dict = None
            self._adjF2C : dict = None
            
            self._adjC2E : dict = None
            self._adjE2C : dict = None
        
        def clear(self):
            super().clear()
            self._adjC2C : Attribute = None
            self._adjV2C : dict = None
            
            self._adjC2F : dict = None
            self._adjF2C : dict = None

            self._adjE2F : dict = None
            self._adjE2C : dict = None
            self._adjC2E : dict = None

        def _compute_connectivity(self):
            super()._compute_connectivity()
            
            self._adjV2C = dict([(i,set()) for i in self.mesh.id_vertices ])
            for iC,C in enumerate(self.mesh.cells):
                for V in C:
                    self._adjV2C[V].add(iC)

            for U in self.mesh.id_vertices:
                self._adjV2C[U] = list(self._adjV2C[U])

        def _compute_adjacent_cell(self):
            # GEOGRAM_API CellDescriptor tet_descriptor = {
            #     4,         // nb_vertices
            #     4,         // nb_faces
            #     {3,3,3,3}, // nb_vertices in face
            #     {          // faces
            #         {1,3,2},
            #         {0,2,3},
            #         {3,1,0},
            #         {0,1,2}
            #     },
            #     6,         // nb_edges
            #     {          // edges
            #         {1,2}, {2,3}, {3,1}, {0,1}, {0,2}, {0,3}
            #     },
            #     {          // edges adjacent faces
            #         {0,3}, {0,1}, {0,2}, {2,3}, {3,1}, {1,2}
            #     }
            # };
            if self.mesh.cell_faces.has_attribute("adjacent_cell"):
                self._adjC2C = self.mesh.cell_faces.get_attribute("adjacent_cell")
            else:
                self._adjC2C = self.mesh.cell_faces.create_attribute("adjacent_cell", int, 1, default_value= config.NOT_AN_ID)
                for iC, cell in enumerate(self.mesh.cells):
                    v0,v1,v2,v3 = cell
                    # face fi does not contain vertex vi
                    f0,f1,f2,f3 = self.face_id(v1,v3,v2), self.face_id(v0,v2,v3), self.face_id(v3,v1,v0), self.face_id(v0,v1,v2)
                    for iF, F in enumerate((f0,f1,f2,f3)):
                        for iC2 in self.face_to_cells(F):
                            if iC2 != iC: self._adjC2C[(iC,iF)] = iC2
                            

        def _compute_cell_adj(self):
            self._adjC2F = dict([(i,[]) for i in self.mesh.id_cells ])
            self._adjF2C = dict([(i,[]) for i in self.mesh.id_faces ])

            for iC,C in enumerate(self.mesh.cells):
                if len(C)==4: # tetrahedra : every subset of 3 elements is a face
                    for i in range(4):
                        # By convention, ith adjacent face does not contain vertex i of the tet
                        F = C[:i] + C[i+1:]
                        iF = self.face_id(*F)
                        self._adjC2F[iC].append(iF)
                        self._adjF2C[iF].append(iC)                        
                else: # not all subsets are faces -> more complicated
                    self._adjC2F[iC] = set()
                    for iF,F in enumerate(self.mesh.faces):
                            sF = set(F)
                            sC = set(C)
                            if sC.issuperset(sF):
                                self._adjC2F[iC].add(iF)
                                self._adjF2C[iF].append(iC)
            for iC in self.mesh.id_cells:
                self._adjC2F[iC] = list(self._adjC2F[iC])
            for iF in self.mesh.id_faces:
                self._adjF2C[iF] = list(self._adjF2C[iF])

        def _compute_edge_id(self):
            super()._compute_edge_id() # for edge_id computation
            self._adjE2F = dict([(e, []) for e in self.mesh.id_edges])
            self._adjE2C = dict([(e, set()) for e in self.mesh.id_edges])

            for f in self.mesh.id_faces:
                c = set(self.face_to_cells(f))
                for e in self.face_to_edges(f):
                    self._adjE2F[e].append(f)
                    self._adjE2C[e] |= c
            for e in self.mesh.id_edges:
                self._adjE2C[e] = list(self._adjE2C[e])
            
            if config.sort_neighborhoods:
                self._sort_edge_neighborhoods()

        def _sort_edge_neighborhoods(self):
            """
            Sorts cells adjacent to an edge so that they are in direct order
            """
            if not self.mesh.is_tetrahedral(): return # only makes sense for tet meshes
            for e,(A,B) in enumerate(self.mesh.edges):
                keys_cell = dict()
                keys_face = dict()
                kc = 0
                kf = 0
                iC = self._adjE2C[e][0]
                keys_cell[iC] = kc
                p1,p2 = (x for x in self.mesh.cells[iC] if x not in (A,B)) # pivots
                while True:
                    face = self.face_id(A,B,p1)
                    kf += 1
                    keys_face[face] = kf
                    nextC = self.other_face_side(iC,face)
                    if nextC is None or nextC in keys_cell: break # we have gone full circle
                    kc += 1
                    keys_cell[nextC] = kc
                    iC = nextC
                    p1 = [x for x in self.mesh.cells[iC] if x not in (A,B,p1)][0]
                
                kc = 0
                kf = 0
                iC = self._adjE2C[e][0]
                while True:
                    face = self.face_id(A,B,p2)
                    kf -= 1
                    keys_face[face] = kf
                    nextC = self.other_face_side(iC,face)
                    if nextC is None or nextC in keys_cell: break # we have gone full circle
                    kc -= 1
                    keys_cell[nextC] = kc
                    iC = nextC
                    p2 = [x for x in self.mesh.cells[iC] if x not in (A,B,p2)][0]
                self._adjE2C[e].sort(key= lambda c : keys_cell[c])
                self._adjE2F[e].sort(key= lambda f : keys_face[f])

        ##### Faces - Cells #####

        def n_F2C(self, F):
            return len(self.face_to_cells(F))

        def face_to_cells(self, iF):
            if self._adjF2C is None:
                self._compute_cell_adj()
            return self._adjF2C[iF]

        def cell_to_face(self, iC):
            if self._adjC2F is None:
                self._compute_cell_adj()
            return self._adjC2F[iC]

        ##### Cell - Cell #####

        def cell_to_cell(self, iC):
            if self._adjC2C is None:
                self._compute_adjacent_cell()
            return [ self._adjC2C[(iC,i)] for i in range(len(self.mesh.cells[iC])) if self._adjC2C[(iC,i)] != config.NOT_AN_ID]

        def other_face_side(self, C, F):
            if len(self.face_to_cells(F)) != 2 : return None
            C1,C2 = self.face_to_cells(F)
            if C==C1: return C2
            if C==C2 : return C1
            return None
        
        def common_face(self, C1, C2):
            common_vert = set(self.mesh.cells[C1]).intersection(self.mesh.cells[C2])
            if len(common_vert) != 3: return None
            return self.face_id(*common_vert)

        ##### Vertex - Cell #####
        def vertex_to_cell(self, V):
            if self._adjV2C is None:
                self._compute_connectivity()
            return self._adjV2C[V]
        
        def cell_to_vertex(self, C):
            return self.mesh.cells[C]
        
        def in_cell_index(self, C, V):
            """Index of vertex V in cell C. None if V is not in cell C

            Parameters:
                C (int): cell index
                V (int): vertex index

            Returns:
                int
            """
            for (i,v) in enumerate(self.mesh.cells[C]):
                if v==V: return i
            return None

        def in_cell_face_index(self,C,F):
            face_set = set(self.mesh.faces[F])
            for i,_ in enumerate(self.mesh.cells[C]):
                cell_set_i = set(self.mesh.cells[C][:i] + self.mesh.cells[C][(i+1):])
                if face_set == cell_set_i:
                    return i
            return None

        ##### Edge - face #####

        # face_to_edge already defined in parent class SurfaceMesh.connectivity
        # edge_to_face is not since SurfaceMesh takes advantage of the half edge structure
        
        def edge_to_face(self, e):
            if self._adjE2F is None:
                self._compute_edge_id()
            return self._adjE2F[e]

        ##### Edge - Cells #####

        def cell_to_edge(self, c):
            if self._adjC2E is None:
                self._adjC2E = dict()
            if self._adjC2E.get(c, None) is None:
                # build connectivity for cell C
                self._adjC2E[c] = []
                verts = self.mesh.cells[c]
                for i in range(len(verts)):
                    for j in range(i):
                        A,B = verts[i], verts[j]
                        e = self.edge_id(A,B)
                        if e is not None:
                            self._adjC2E[c].append(e)
            return self._adjC2E[c]

        def edge_to_cell(self,e):
            if self._adjE2C is None:
                self._compute_edge_id()
            return self._adjE2C[e]

    class _BoundaryConnectivity(SurfaceMesh._Connectivity):

        def __init__(self, master):
            """
            Attributes:
                m2b_vertex (dict): volume mesh -> boundary indirection for vertices
                m2b_edge (dict): volume mesh -> boundary indirection for edges
                m2b_face (dict): volume mesh -> boundary indirection for faces

                b2m_vertex (dict): boundary -> volume mesh indirection for vertices
                b2m_edge (dict): boundary -> volume mesh indirection for edges
                b2m_face (dict): boundary -> volume mesh indirection for faces

            """
            self.complete_mesh = master

            # indirection maps between boundary elements and their indices in the full mesh
            self.m2b_vertex : dict = None # mesh to boundary
            self.m2b_edge : dict = None
            self.m2b_face : dict = None

            self.b2m_vertex : dict = None # boundary to mesh
            self.b2m_edge : dict = None
            self.b2m_face : dict = None

            mesh = self._extract_surface_boundary() # builds indirection maps for vertices and faces
            super().__init__(mesh) # the boundary mesh can be accessed using the .master attribute

            # build indirection map for edges
            for e in self.complete_mesh.boundary_edges:
                u,v = self.complete_mesh.edges[e]
                bu,bv = self.m2b_vertex[u], self.m2b_vertex[v]
                be = self.edge_id(bu,bv)
                self.m2b_edge[e] = be
                self.b2m_edge[be] = e            

        def _extract_surface_boundary(self):
            """
            Builds indirection maps. This is performed automatically upon first connectivity request.
            """
            boundary = RawMeshData()
            # indirection maps
            self.m2b_vertex, self.b2m_vertex = dict(), dict()
            self.m2b_edge, self.b2m_edge = dict(), dict()
            self.m2b_face, self.b2m_face = dict(), dict()
            vertex_set = set()
            bnd_faces = []
            for i,iF in enumerate(self.complete_mesh.boundary_faces):
                self.m2b_face[iF] = i
                self.b2m_face[i] = iF
                bnd_faces.append(iF)
                for v in self.complete_mesh.faces[iF]:
                    vertex_set.add(v)

            # re order vertices
            for i,v in enumerate(vertex_set):
                boundary.vertices.append(self.complete_mesh.vertices[v])
                self.m2b_vertex[v] = i
                self.b2m_vertex[i] = v

            # generate set of faces and check for orientation
            for iF in bnd_faces:
                iC = self.complete_mesh.connectivity.face_to_cells(iF)[0]
                pA,pB,pC = (self.complete_mesh.vertices[_x] for _x in self.complete_mesh.faces[iF])
                bA,bB,bC = (self.m2b_vertex[v] for v in self.complete_mesh.faces[iF])
                D = [x for x in self.complete_mesh.cells[iC] if x not in self.complete_mesh.faces[iF]][0] # fourth point in cell but not on face
                pD = self.complete_mesh.vertices[D]
                if det_3x3(pA-pD,pB-pD,pC-pD)>0:
                    boundary.faces.append((bA,bB,bC))
                else:
                    boundary.faces.append((bA,bC,bB))
            return SurfaceMesh(boundary)
        
        ##### Vertex to elements #####

        def vertex_to_vertices(self, V : int):
            Vb = self.m2b_vertex.get(V,None)
            if Vb is None: return []
            return [self.b2m_vertex[_v] for _v in super().vertex_to_vertices(Vb)]
        
        def vertex_to_edges(self, V : int) :
            Vb = self.m2b_vertex.get(V,None)
            if Vb is None: return []
            edges = [self.edge_id(Vb,Wb) for Wb in super().vertex_to_vertices(Vb)]
            return [self.b2m_edge[e] for e in edges]

        def vertex_to_faces(self, V):
            Vb = self.m2b_vertex.get(V,None)
            if Vb is None: return
            return [self.b2m_face[f] for f in super().vertex_to_faces(Vb)]

        def vertex_to_face_quad(self, V):
            raise NotImplementedError
            
        ##### Faces to elements #####

        def face_to_vertices(self, F):
            Fb = self.m2b_face.get(F,None)
            if Fb is None : return []
            return list(self.mesh.faces[Fb])

        def in_face_index(self, F, V):
            """Index of vertex V in face F. None if V is not in face F

            Parameters:
                F (int): face index
                V (int): vertex index

            Returns:
                int
            """
            Fb,Vb = self.m2b_face.get(F,None), self.m2b_vertex.get(V,None)
            if Fb is None or Vb is None : return None
            return super().in_face_index(Fb,Vb)
        
        def face_to_edges(self, F):
            Fb = self.m2b_face.get(F,None)
            if Fb is None: return []
            return  [self.b2m_edge[_e] for _e in super().face_to_edges(Fb)]

        def face_to_faces(self, F):
            Fb = self.m2b_face.get(F,None)
            if Fb is None: return []
            return [self.b2m_face[_f] for _f in super().face_to_faces(Fb)]